# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Application and entities that are used to describe Users on Discord."""

from __future__ import annotations

__all__: typing.Sequence[str] = ("OwnUser", "PartialUser", "PremiumType", "User", "UserFlag")

import abc
import typing

import attrs

from hikari import snowflakes
from hikari import traits
from hikari import undefined
from hikari import urls
from hikari.internal import attrs_extensions
from hikari.internal import deprecation
from hikari.internal import enums
from hikari.internal import routes

if not typing.TYPE_CHECKING:
    # This is insanely hacky, but it is needed for ruff to not complain until it gets type inference
    from hikari.internal import typing_extensions

if typing.TYPE_CHECKING:
    import datetime

    import typing_extensions  # noqa: TC004

    from hikari import channels
    from hikari import colors
    from hikari import embeds as embeds_
    from hikari import files
    from hikari import guilds
    from hikari import locales
    from hikari import messages
    from hikari.api import special_endpoints


@typing.final
class UserFlag(enums.Flag):
    """The known user flags that represent account badges."""

    NONE = 0
    """None."""

    DISCORD_EMPLOYEE = 1 << 0
    """Discord Employee."""

    PARTNERED_SERVER_OWNER = 1 << 1
    """Partnered Discord server owner."""

    HYPESQUAD_EVENTS = 1 << 2
    """Participated in HypeSquad Events."""

    BUG_HUNTER_LEVEL_1 = 1 << 3
    """Participated in the Discord Testers community."""

    HYPESQUAD_BRAVERY = 1 << 6
    """Member of the HypeSquad House of Bravery."""

    HYPESQUAD_BRILLIANCE = 1 << 7
    """Member of the HypeSquad House of Brilliance."""

    HYPESQUAD_BALANCE = 1 << 8
    """Member of the HypeSquad House of Balance."""

    EARLY_SUPPORTER = 1 << 9
    """Early premium supporter of Discord."""

    TEAM_USER = 1 << 10
    """Pseudo-user for an app's development team."""

    BUG_HUNTER_LEVEL_2 = 1 << 14
    """Vastly participated in the Discord Testers community."""

    VERIFIED_BOT = 1 << 16
    """Verified bot."""

    EARLY_VERIFIED_DEVELOPER = 1 << 17
    """Early verified bot developer."""

    DISCORD_CERTIFIED_MODERATOR = 1 << 18
    """Alumni of the moderator program."""

    BOT_HTTP_INTERACTIONS = 1 << 19
    """Uses only HTTP interactions and is shown in the active member list."""

    SPAMMER = 1 << 20
    """Suspected of being a spammer."""

    ACTIVE_DEVELOPER = 1 << 22
    """Active developer of a Discord app."""

    PROVISIONAL_ACCOUNT = 1 << 23
    """Provisional account used with the social layer integration."""

    QUARANTINED = 1 << 44
    """Account is quarantined."""

    COLLABORATOR = 1 << 50
    """Discord Collaborator, considered Staff."""

    RESTRICTED_COLLABORATOR = 1 << 51
    """Restricted Discord Collaborator, considered Staff."""


@typing.final
class PremiumType(int, enums.Enum):
    """The types of Nitro."""

    NONE = 0
    """No premium."""

    NITRO_CLASSIC = 1
    """Legacy premium tier, including basic perks like animated emojis and avatars."""

    NITRO = 2
    """Premium tier including *all* available perks (e.g. 2 server boosts)."""

    NITRO_BASIC = 3
    """Premium tier including basic perks (e.g. animated emojis and avatars)."""


@attrs.define(kw_only=True, weakref_slot=False)
class AvatarDecoration:
    """Data for an avatar decoration."""

    asset_hash: str = attrs.field(repr=True)
    """The hash of the asset."""

    sku_id: snowflakes.Snowflake = attrs.field(repr=True)
    """The ID of the asset's SKU."""

    expires_at: datetime.datetime | None = attrs.field(repr=True)
    """The datetime at which the user will no longer have access to the avatar decoration."""

    @property
    @deprecation.deprecated("Use 'make_url' instead.")
    def url(self) -> files.URL:
        """Default image URL for this avatar decoration."""
        deprecation.warn_deprecated("url", removal_version="2.5.0", additional_info="Use 'make_url' instead.")
        return self.make_url()

    def make_url(
        self,
        *,
        file_format: undefined.UndefinedOr[typing.Literal["PNG", "JPEG", "JPG", "WEBP", "APNG"]] = undefined.UNDEFINED,
        size: int = 4096,
        lossless: bool = True,
    ) -> files.URL:
        """Generate the image URL for this avatar decoration.

        Parameters
        ----------
        file_format
            The format to use for this URL.

            Supports `PNG`, `JPEG`, `JPG`, `WEBP`, and `APNG`.

            If not specified, the format will be `PNG` or `APNG` based on
            whether the decoration is animated or not.
        size
            The size to set for the URL;
            This is ignored for the `APNG` format;
            Can be any power of two between `16` and `4096`.
        lossless
            Whether to return a lossless or compressed WEBP image;
            This is ignored if `file_format` is not `WEBP`.

        Returns
        -------
        hikari.files.URL
            The URL to the avatar decoration.

        Raises
        ------
        TypeError
            If an invalid format is passed for `file_format`;
            If an animated format is requested for a static avatar decoration.
        ValueError
            If `size` is specified but is not a power of two or not between 16 and 4096.
        """
        animated = self.asset_hash.startswith("a_")

        if not file_format:
            file_format = "APNG" if animated else "PNG"

        return routes.CDN_AVATAR_DECORATION.compile_to_file(
            urls.MEDIA_PROXY_URL, hash=self.asset_hash, size=size, file_format=file_format, lossless=lossless
        )


class PartialUser(snowflakes.Unique, abc.ABC):
    """A partial interface for a user.

    Fields may or may not be present, and should be explicitly checked
    before using them to ensure they are not [`hikari.undefined.UNDEFINED`][].

    This is used for endpoints and events that only expose partial user
    information.

    For full user info, consider calling the [`hikari.users.PartialUser.fetch_self`][] method to perform an
    API call.
    """

    __slots__: typing.Sequence[str] = ()

    @property
    @abc.abstractmethod
    def app(self) -> traits.RESTAware:
        """Client application that models may use for procedures."""

    @property
    @abc.abstractmethod
    def avatar_decoration(self) -> undefined.UndefinedNoneOr[AvatarDecoration]:
        """Avatar decoration for the user, if they have one, otherwise [`None`][]."""

    @property
    @abc.abstractmethod
    def avatar_hash(self) -> undefined.UndefinedNoneOr[str]:
        """Avatar hash for the user, if they have one, otherwise [`None`][]."""

    @property
    @abc.abstractmethod
    def banner_hash(self) -> undefined.UndefinedNoneOr[str]:
        """Banner hash for the user, if they have one, otherwise [`None`][]."""

    @property
    @abc.abstractmethod
    def accent_color(self) -> undefined.UndefinedNoneOr[colors.Color]:
        """Custom banner color for the user if set, else [`None`][].

        The official client will decide the default color if not set.
        """

    @property
    def accent_colour(self) -> undefined.UndefinedNoneOr[colors.Color]:
        """Alias for the [`hikari.users.PartialUser.accent_color`][] field."""
        return self.accent_color

    @property
    @abc.abstractmethod
    def discriminator(self) -> undefined.UndefinedOr[str]:
        """Discriminator for the user.

        !!! deprecated 2.0.0.dev120
            Discriminators are deprecated and being replaced with "0" by Discord
            during username migration. This field will be removed after migration is complete.
            Learn more here: https://dis.gd/usernames
        """

    @property
    @abc.abstractmethod
    def username(self) -> undefined.UndefinedOr[str]:
        """Username for the user."""

    @property
    @abc.abstractmethod
    def global_name(self) -> undefined.UndefinedNoneOr[str]:
        """Global name for the user, if they have one, otherwise [`None`][]."""

    @property
    def display_name(self) -> undefined.UndefinedNoneOr[str]:
        """Return the user's display name.

        Either users global name (if set) or its username.
        """
        return self.global_name or self.username

    @property
    @abc.abstractmethod
    def is_bot(self) -> undefined.UndefinedOr[bool]:
        """Whether this user is a bot account."""

    @property
    @abc.abstractmethod
    def is_system(self) -> undefined.UndefinedOr[bool]:
        """Whether  this user is a system account."""

    @property
    @abc.abstractmethod
    def flags(self) -> undefined.UndefinedOr[UserFlag]:
        """Flag bits that are set for the user."""

    @property
    @abc.abstractmethod
    def mention(self) -> str:
        """Return a raw mention string for the given user.

        Examples
        --------
        ```py
        >>> some_user.mention
        '<@123456789123456789>'
        ```
        """

    async def fetch_dm_channel(self) -> channels.DMChannel:
        """Fetch the DM channel for this user.

        Returns
        -------
        hikari.channels.DMChannel
            The requested channel.

        Raises
        ------
        hikari.errors.UnauthorizedError
            If you are unauthorized to make the request (invalid/missing token).
        hikari.errors.NotFoundError
            If the user is not found.
        hikari.errors.RateLimitTooLongError
            Raised in the event that a rate limit occurs that is
            longer than `max_rate_limit` when making a request.
        hikari.errors.InternalServerError
            If an internal error occurs on Discord while handling the request.
        """
        return await self.app.rest.create_dm_channel(self.id)

    async def fetch_self(self) -> User:
        """Get this user's up-to-date object by performing an API call.

        Returns
        -------
        hikari.users.User
            The requested user object.

        Raises
        ------
        hikari.errors.NotFoundError
            If the user is not found.
        hikari.errors.RateLimitTooLongError
            Raised in the event that a rate limit occurs that is
            longer than `max_rate_limit` when making a request.
        hikari.errors.InternalServerError
            If an internal error occurs on Discord while handling the request.
        """
        return await self.app.rest.fetch_user(user=self.id)

    async def send(
        self,
        content: undefined.UndefinedOr[typing.Any] = undefined.UNDEFINED,
        *,
        attachment: undefined.UndefinedOr[files.Resourceish] = undefined.UNDEFINED,
        attachments: undefined.UndefinedOr[typing.Sequence[files.Resourceish]] = undefined.UNDEFINED,
        component: undefined.UndefinedOr[special_endpoints.ComponentBuilder] = undefined.UNDEFINED,
        components: undefined.UndefinedOr[typing.Sequence[special_endpoints.ComponentBuilder]] = undefined.UNDEFINED,
        embed: undefined.UndefinedOr[embeds_.Embed] = undefined.UNDEFINED,
        embeds: undefined.UndefinedOr[typing.Sequence[embeds_.Embed]] = undefined.UNDEFINED,
        tts: undefined.UndefinedOr[bool] = undefined.UNDEFINED,
        reply: undefined.UndefinedOr[snowflakes.SnowflakeishOr[messages.PartialMessage]] = undefined.UNDEFINED,
        reply_must_exist: undefined.UndefinedOr[bool] = undefined.UNDEFINED,
        mentions_everyone: undefined.UndefinedOr[bool] = undefined.UNDEFINED,
        mentions_reply: undefined.UndefinedOr[bool] = undefined.UNDEFINED,
        user_mentions: undefined.UndefinedOr[snowflakes.SnowflakeishSequence[PartialUser] | bool] = undefined.UNDEFINED,
        role_mentions: undefined.UndefinedOr[
            snowflakes.SnowflakeishSequence[guilds.PartialRole] | bool
        ] = undefined.UNDEFINED,
        flags: undefined.UndefinedType | int | messages.MessageFlag = undefined.UNDEFINED,
    ) -> messages.Message:
        """Send a message to this user in DM's.

        Parameters
        ----------
        content
            If provided, the message contents. If
            [`hikari.undefined.UNDEFINED`][], then nothing will be sent
            in the content. Any other value here will be cast to a
            [`str`][].

            If this is a [`hikari.embeds.Embed`][] and no `embed` nor `embeds` kwarg
            is provided, then this will instead update the embed. This allows
            for simpler syntax when sending an embed alone.

            Likewise, if this is a [`hikari.files.Resource`][], then the
            content is instead treated as an attachment if no `attachment` and
            no `attachments` kwargs are provided.
        attachment
            If provided, the message attachment. This can be a resource,
            or string of a path on your computer or a URL.

            Attachments can be passed as many different things, to aid in
            convenience.

            - If a [`pathlib.PurePath`][] or [`str`][] to a valid URL, the
                resource at the given URL will be streamed to Discord when
                sending the message. Subclasses of
                [`hikari.files.WebResource`][] such as
                [`hikari.files.URL`][],
                [`hikari.messages.Attachment`][],
                [`hikari.emojis.Emoji`][],
                [`hikari.embeds.EmbedResource`][], etc will also be uploaded this way.
                This will use bit-inception, so only a small percentage of the
                resource will remain in memory at any one time, thus aiding in
                scalability.
            - If a [`hikari.files.Bytes`][] is passed, or a [`str`][]
                that contains a valid data URI is passed, then this is uploaded
                with a randomized file name if not provided.
            - If a [`hikari.files.File`][], [`pathlib.PurePath`][] or
                [`str`][] that is an absolute or relative path to a file
                on your file system is passed, then this resource is uploaded
                as an attachment using non-blocking code internally and streamed
                using bit-inception where possible. This depends on the
                type of [`concurrent.futures.Executor`][] that is being used for
                the application (default is a thread pool which supports this
                behaviour).
        attachments
            If provided, the message attachments. These can be resources, or
            strings consisting of paths on your computer or URLs.
        component
            If provided, builder object of the component to include in this message.
        components
            If provided, a sequence of the component builder objects to include
            in this message.
        embed
            If provided, the message embed.
        embeds
            If provided, the message embeds.
        tts
            If provided, whether the message will be read out by a screen
            reader using Discord's TTS (text-to-speech) system.
        reply
            If provided, the message to reply to.
        reply_must_exist
            If provided, whether to error if the message being replied to does
            not exist instead of sending as a normal (non-reply) message.

            This will not do anything if not being used with `reply`.
        mentions_everyone
            If provided, whether the message should parse @everyone/@here
            mentions.
        mentions_reply
            If provided, whether to mention the author of the message
            that is being replied to.

            This will not do anything if not being used with `reply`.
        user_mentions
            If provided, and [`True`][], all user mentions will be detected.
            If provided, and [`False`][], all user mentions will be ignored
            if appearing in the message body.
            Alternatively this may be a collection of
            [`hikari.snowflakes.Snowflake`][], or
            [`hikari.users.PartialUser`][] derivatives to enforce mentioning
            specific users.
        role_mentions
            If provided, and [`True`][], all role mentions will be detected.
            If provided, and [`False`][], all role mentions will be ignored
            if appearing in the message body.
            Alternatively this may be a collection of
            [`hikari.snowflakes.Snowflake`][], or
            [`hikari.guilds.PartialRole`][] derivatives to enforce mentioning
            specific roles.
        flags
            If provided, optional flags to set on the message. If
            [`hikari.undefined.UNDEFINED`][], then nothing is changed.

            Note that some flags may not be able to be set. Currently the only
            flags that can be set are [`hikari.messages.MessageFlag.NONE`][]
            and [`hikari.messages.MessageFlag.SUPPRESS_EMBEDS`][].

        Returns
        -------
        hikari.messages.Message
            The created message.

        Raises
        ------
        ValueError
            If more than 100 unique objects/entities are passed for
            `role_mentions` or `user_mentions`.
        TypeError
            If both `attachment` and `attachments` are specified.
        hikari.errors.BadRequestError
            This may be raised in several discrete situations, such as messages
            being empty with no attachments or embeds; messages with more than
            2000 characters in them, embeds that exceed one of the many embed
            limits; too many attachments; attachments that are too large;
            invalid image URLs in embeds; `reply` not found or not in the same
            channel; too many components.
        hikari.errors.UnauthorizedError
            If you are unauthorized to make the request (invalid/missing token).
        hikari.errors.ForbiddenError
            If you are missing the [`hikari.permissions.Permissions.SEND_MESSAGES`][] in
            the channel or the person you are trying to message has the DM's disabled.
        hikari.errors.NotFoundError
            If the user is not found.
        hikari.errors.RateLimitTooLongError
            Raised in the event that a rate limit occurs that is
            longer than `max_rate_limit` when making a request.
        hikari.errors.InternalServerError
            If an internal error occurs on Discord while handling the request.
        """
        channel_id = None
        if isinstance(self.app, traits.CacheAware):
            channel_id = self.app.cache.get_dm_channel_id(self.id)

        if channel_id is None:
            channel_id = (await self.fetch_dm_channel()).id

        return await self.app.rest.create_message(
            channel=channel_id,
            content=content,
            attachment=attachment,
            attachments=attachments,
            component=component,
            components=components,
            embed=embed,
            embeds=embeds,
            tts=tts,
            reply=reply,
            reply_must_exist=reply_must_exist,
            mentions_everyone=mentions_everyone,
            user_mentions=user_mentions,
            role_mentions=role_mentions,
            mentions_reply=mentions_reply,
            flags=flags,
        )


class User(PartialUser, abc.ABC):
    """Interface for any user-like object.

    This does not include partial users, as they may not be fully formed.
    """

    __slots__: typing.Sequence[str] = ()

    @property
    @abc.abstractmethod
    @typing_extensions.override
    def app(self) -> traits.RESTAware:
        """Client application that models may use for procedures."""

    @property
    @abc.abstractmethod
    @typing_extensions.override
    def accent_color(self) -> colors.Color | None:
        """The custom banner color for the user, if set else [`None`][].

        The official client will decide the default color if not set.
        """

    @property
    @typing_extensions.override
    def accent_colour(self) -> colors.Color | None:
        """Alias for the [`hikari.users.User.accent_color`][] field."""
        return self.accent_color

    @property
    @abc.abstractmethod
    @typing_extensions.override
    def avatar_decoration(self) -> AvatarDecoration | None:
        """Avatar decoration for the user, if they have one, otherwise [`None`][]."""

    @property
    @abc.abstractmethod
    @typing_extensions.override
    def avatar_hash(self) -> str | None:
        """Avatar hash for the user, if they have one, otherwise [`None`][]."""

    @property
    @deprecation.deprecated("Use 'make_avatar_url' instead.")
    def avatar_url(self) -> files.URL | None:
        """Avatar URL for the user, if they have one set.

        Will be [`None`][] if no avatar is set. In this case, you
        should use [`hikari.User.default_avatar_url`][] instead.
        """
        deprecation.warn_deprecated(
            "avatar_url", removal_version="2.5.0", additional_info="Use 'make_avatar_url' instead."
        )
        return self.make_avatar_url()

    @property
    @abc.abstractmethod
    @typing_extensions.override
    def banner_hash(self) -> str | None:
        """Banner hash for the user, if they have one, otherwise [`None`][]."""

    @property
    @deprecation.deprecated("Use 'make_banner_url' instead.")
    def banner_url(self) -> files.URL | None:
        """Banner URL for the user, if they have one set.

        Will be [`None`][] if no banner is set.
        """
        deprecation.warn_deprecated(
            "banner_url", removal_version="2.5.0", additional_info="Use 'make_banner_url' instead."
        )
        return self.make_banner_url()

    @property
    def default_avatar_url(self) -> files.URL:
        """Default avatar URL for this user."""
        if self.discriminator == "0":  # migrated account
            return routes.CDN_DEFAULT_USER_AVATAR.compile_to_file(
                urls.CDN_URL, style=(self.id >> 22) % 6, file_format="PNG"
            )

        return routes.CDN_DEFAULT_USER_AVATAR.compile_to_file(
            urls.CDN_URL, style=int(self.discriminator) % 5, file_format="PNG"
        )

    @property
    def display_avatar_decoration(self) -> AvatarDecoration | None:
        """Display avatar decoration for the user, if they have one set.

        Will be [`None`][] if no avatar decoration is set.
        """
        return self.avatar_decoration

    @property
    def display_avatar_url(self) -> files.URL:
        """Display avatar URL for this user."""
        return self.make_avatar_url() or self.default_avatar_url

    @property
    def display_banner_url(self) -> files.URL | None:
        """Display banner URL for this user, if they have one set.

        Will be [`None`][] if no banner is set.
        """
        return self.make_banner_url()

    @property
    @abc.abstractmethod
    @typing_extensions.override
    def discriminator(self) -> str:
        """Discriminator for the user.

        !!! deprecated 2.0.0.dev120
            Discriminators are deprecated and being replaced with "0" by Discord
            during username migration. This field will be removed after migration is complete.
            Learn more here: https://dis.gd/usernames
        """

    @property
    @abc.abstractmethod
    @typing_extensions.override
    def flags(self) -> UserFlag:
        """Flag bits that are set for the user."""

    @property
    @abc.abstractmethod
    @typing_extensions.override
    def is_bot(self) -> bool:
        """Whether this user is a bot account."""

    @property
    @abc.abstractmethod
    @typing_extensions.override
    def is_system(self) -> bool:
        """Whether this user is a system account."""

    @property
    @abc.abstractmethod
    @typing_extensions.override
    def mention(self) -> str:
        """Return a raw mention string for the given user.

        Examples
        --------
        ```py
        >>> some_user.mention
        '<@123456789123456789>'
        ```
        """

    @property
    @abc.abstractmethod
    @typing_extensions.override
    def username(self) -> str:
        """Username for the user."""

    @property
    @abc.abstractmethod
    @typing_extensions.override
    def global_name(self) -> str | None:
        """Global name for the user, if they have one, otherwise [`None`][]."""

    def make_avatar_url(
        self,
        *,
        file_format: undefined.UndefinedOr[
            typing.Literal["PNG", "JPEG", "JPG", "WEBP", "AWEBP", "GIF"]
        ] = undefined.UNDEFINED,
        size: int = 4096,
        lossless: bool = True,
        ext: str | None | undefined.UndefinedType = undefined.UNDEFINED,
    ) -> files.URL | None:
        """Generate the avatar URL for this user, if set.

        If no avatar is set, this returns [`None`][].

        Parameters
        ----------
        file_format
            The format to use for this URL.

            Supports `PNG`, `JPEG`, `JPG`, `WEBP`, `AWEBP` and `GIF`.

            If not specified, the format will be determined based on
            whether the avatar is animated or not.
        size
            The size to set for the URL;
            Can be any power of two between `16` and `4096`;
        lossless
            Whether to return a lossless or compressed WEBP image;
            This is ignored if `file_format` is not `WEBP` or `AWEBP`.
        ext
            The ext to use for this URL.
            Supports `png`, `jpeg`, `jpg`, `webp` and `gif` (when
            animated).

            If [`None`][], then the correct default extension is
            determined based on whether the avatar is animated or not.

            !!! deprecated 2.4.0
                This has been replaced with the `file_format` argument.

        Returns
        -------
        typing.Optional[hikari.files.URL]
            The URL, or [`None`][] if no avatar is set.

        Raises
        ------
        TypeError
            If an invalid format is passed for `file_format`;
            If an animated format is requested for a static avatar.
        ValueError
            If `size` is specified but is not a power of two or not between 16 and 4096.
        """
        if self.avatar_hash is None:
            return None

        if ext:
            deprecation.warn_deprecated(
                "ext", removal_version="2.5.0", additional_info="Use 'file_format' argument instead."
            )
            file_format = ext.upper()  # type: ignore[assignment]

        if not file_format:
            file_format = "GIF" if self.avatar_hash.startswith("a_") else "PNG"

        return routes.CDN_USER_AVATAR.compile_to_file(
            urls.CDN_URL, user_id=self.id, hash=self.avatar_hash, size=size, file_format=file_format, lossless=lossless
        )

    def make_banner_url(
        self,
        *,
        file_format: undefined.UndefinedOr[
            typing.Literal["PNG", "JPEG", "JPG", "WEBP", "AWEBP", "GIF"]
        ] = undefined.UNDEFINED,
        size: int = 4096,
        lossless: bool = True,
        ext: str | None | undefined.UndefinedType = undefined.UNDEFINED,
    ) -> files.URL | None:
        """Generate the banner URL for this user, if set.

        If no banner is set, this returns [`None`][].

        Parameters
        ----------
        file_format
            The format to use for this URL.

            Supports `PNG`, `JPEG`, `JPG`, `WEBP`, `AWEBP` and `GIF`.

            If not specified, the format will be determined based on
            whether the banner is animated or not.
        size
            The size to set for the URL;
            Can be any power of two between `16` and `4096`;
        lossless
            Whether to return a lossless or compressed WEBP image;
            This is ignored if `file_format` is not `WEBP` or `AWEBP`.
        ext
            The ext to use for this URL.
            Supports `png`, `jpeg`, `jpg`, `webp` and `gif` (when
            animated).

            If [`None`][], then the correct default extension is
            determined based on whether the banner is animated or not.

            !!! deprecated 2.4.0
                This has been replaced with the `file_format` argument.

        Returns
        -------
        typing.Optional[hikari.files.URL]
            The URL, or [`None`][] if no banner is set.

        Raises
        ------
        TypeError
            If an invalid format is passed for `file_format`;
            If an animated format is requested for a static banner.
        ValueError
            If `size` is specified but is not a power of two or not between 16 and 4096.
        """
        if self.banner_hash is None:
            return None

        if ext:
            deprecation.warn_deprecated(
                "ext", removal_version="2.5.0", additional_info="Use 'file_format' argument instead."
            )
            file_format = ext.upper()  # type: ignore[assignment]

        if not file_format:
            file_format = "GIF" if self.banner_hash.startswith("a_") else "PNG"

        return routes.CDN_USER_BANNER.compile_to_file(
            urls.CDN_URL, user_id=self.id, hash=self.banner_hash, size=size, file_format=file_format, lossless=lossless
        )


@attrs_extensions.with_copy
@attrs.define(unsafe_hash=True, kw_only=True, weakref_slot=False)
class PartialUserImpl(PartialUser):
    """Implementation for partial information about a user.

    This is pretty much the same as a normal user, but information may not be
    present, which will be denoted by [`hikari.undefined.UNDEFINED`][].
    """

    id: snowflakes.Snowflake = attrs.field(hash=True, repr=True)
    """The ID of this user."""

    app: traits.RESTAware = attrs.field(
        repr=False, eq=False, hash=False, metadata={attrs_extensions.SKIP_DEEP_COPY: True}
    )
    """Client application that models may use for procedures."""

    discriminator: undefined.UndefinedOr[str] = attrs.field(eq=False, hash=False, repr=True)
    """Four-digit discriminator for the user if unmigrated.

    !!! deprecated 2.0.0.dev120
        Discriminators are deprecated and being replaced with "0" by Discord
        during username migration. This field will be removed after migration is complete.
        Learn more here: https://dis.gd/usernames
    """

    username: undefined.UndefinedOr[str] = attrs.field(eq=False, hash=False, repr=True)
    """Username of the user."""

    global_name: undefined.UndefinedNoneOr[str] = attrs.field(eq=False, hash=False, repr=True)
    """Global name of the user."""

    avatar_decoration: undefined.UndefinedNoneOr[AvatarDecoration] = attrs.field(eq=False, hash=False, repr=False)
    """Avatar decoration of the user, if an avatar decoration is set."""

    avatar_hash: undefined.UndefinedNoneOr[str] = attrs.field(eq=False, hash=False, repr=False)
    """Avatar hash of the user, if a custom avatar is set."""

    banner_hash: undefined.UndefinedNoneOr[str] = attrs.field(eq=False, hash=False, repr=False)
    """Banner hash of the user, if a custom banner is set."""

    accent_color: undefined.UndefinedNoneOr[colors.Color] = attrs.field(eq=False, hash=False, repr=False)
    """The custom banner color for the user, if set.

    The official client will decide the default color if not set.
    """

    is_bot: undefined.UndefinedOr[bool] = attrs.field(eq=False, hash=False, repr=True)
    """Whether this user is a bot account."""

    is_system: undefined.UndefinedOr[bool] = attrs.field(eq=False, hash=False, repr=True)
    """Whether this user is a system account."""

    flags: undefined.UndefinedOr[UserFlag] = attrs.field(eq=False, hash=False, repr=True)
    """Public flags for this user."""

    @property
    @typing_extensions.override
    def mention(self) -> str:
        """Return a raw mention string for the given user.

        Examples
        --------
        ```py
        >>> some_user.mention
        '<@123456789123456789>'
        ```
        """
        return f"<@{self.id}>"

    @typing_extensions.override
    def __str__(self) -> str:
        if self.username is undefined.UNDEFINED or self.discriminator is undefined.UNDEFINED:
            return f"Partial user ID {self.id}"
        if self.discriminator == "0":  # migrated account
            return self.username
        return f"{self.username}#{self.discriminator}"


@attrs.define(unsafe_hash=True, kw_only=True, weakref_slot=False)
class UserImpl(PartialUserImpl, User):
    """Concrete implementation of user information."""

    discriminator: str = attrs.field(eq=False, hash=False, repr=True)
    """The user's discriminator.

    !!! deprecated 2.0.0.dev120
        Discriminators are deprecated and being replaced with "0" by Discord
        during username migration. This field will be removed after migration is complete.
        Learn more here: https://dis.gd/usernames
    """

    username: str = attrs.field(eq=False, hash=False, repr=True)
    """The user's username."""

    global_name: str | None = attrs.field(eq=False, hash=False, repr=True)
    """The user's global name."""

    avatar_decoration: AvatarDecoration | None = attrs.field(eq=False, hash=False, repr=False)
    """Avatar decoration of the user, if they have one."""

    avatar_hash: str | None = attrs.field(eq=False, hash=False, repr=False)
    """The user's avatar hash, if they have one, otherwise [`None`][]."""

    banner_hash: str | None = attrs.field(eq=False, hash=False, repr=False)
    """Banner hash of the user, if they have one, otherwise [`None`][]"""

    accent_color: colors.Color | None = attrs.field(eq=False, hash=False, repr=False)
    """The custom banner color for the user, if set.

    The official client will decide the default color if not set.
    """

    is_bot: bool = attrs.field(eq=False, hash=False, repr=True)
    """[`True`][] if this user is a bot account, [`False`][] otherwise."""

    is_system: bool = attrs.field(eq=False, hash=False, repr=True)
    """[`True`][] if this user is a system account, [`False`][] otherwise."""

    flags: UserFlag = attrs.field(eq=False, hash=False, repr=True)
    """The public flags for this user."""


@attrs.define(unsafe_hash=True, kw_only=True, weakref_slot=False)
class OwnUser(UserImpl):
    """Represents a user with extended OAuth2 information."""

    is_mfa_enabled: bool = attrs.field(eq=False, hash=False, repr=False)
    """Whether the user's account has multi-factor authentication enabled."""

    locale: str | locales.Locale | None = attrs.field(eq=False, hash=False, repr=False)
    """The user's set locale.

    This is not provided in the `READY` event.
    """

    is_verified: bool | None = attrs.field(eq=False, hash=False, repr=False)
    """Whether the email for this user's account has been verified.

    Will be [`None`][] if retrieved through the OAuth2 flow without the `email`
    scope.
    """

    email: str | None = attrs.field(eq=False, hash=False, repr=False)
    """The user's set email.

    Will be [`None`][] if retrieved through OAuth2 flow without the `email`
    scope. Will always be [`None`][] for bot users.
    """

    premium_type: PremiumType | int | None = attrs.field(eq=False, hash=False, repr=False)
    """The type of Nitro Subscription this user account had.

    This will always be [`None`][] for bots.
    """

    @typing_extensions.override
    async def fetch_self(self) -> OwnUser:
        """Get this user's up-to-date object.

        Returns
        -------
        hikari.users.OwnUser
            The requested user object.

        Raises
        ------
        hikari.errors.UnauthorizedError
            If you are unauthorized to make the request (invalid/missing token).
        hikari.errors.RateLimitTooLongError
            Raised in the event that a rate limit occurs that is
            longer than `max_rate_limit` when making a request.
        hikari.errors.InternalServerError
            If an internal error occurs on Discord while handling the request.
        """
        return await self.app.rest.fetch_my_user()

    @typing_extensions.override
    async def fetch_dm_channel(self) -> typing.NoReturn:
        msg = "Unable to fetch your own DM channel"
        raise TypeError(msg)

    @typing_extensions.override
    async def send(
        self,
        content: undefined.UndefinedOr[typing.Any] = undefined.UNDEFINED,
        *,
        attachment: undefined.UndefinedOr[files.Resourceish] = undefined.UNDEFINED,
        attachments: undefined.UndefinedOr[typing.Sequence[files.Resourceish]] = undefined.UNDEFINED,
        component: undefined.UndefinedOr[special_endpoints.ComponentBuilder] = undefined.UNDEFINED,
        components: undefined.UndefinedOr[typing.Sequence[special_endpoints.ComponentBuilder]] = undefined.UNDEFINED,
        embed: undefined.UndefinedOr[embeds_.Embed] = undefined.UNDEFINED,
        embeds: undefined.UndefinedOr[typing.Sequence[embeds_.Embed]] = undefined.UNDEFINED,
        nonce: undefined.UndefinedOr[str] = undefined.UNDEFINED,
        tts: undefined.UndefinedOr[bool] = undefined.UNDEFINED,
        reply: undefined.UndefinedOr[snowflakes.SnowflakeishOr[messages.PartialMessage]] = undefined.UNDEFINED,
        reply_must_exist: undefined.UndefinedOr[bool] = undefined.UNDEFINED,
        mentions_everyone: undefined.UndefinedOr[bool] = undefined.UNDEFINED,
        mentions_reply: undefined.UndefinedOr[bool] = undefined.UNDEFINED,
        user_mentions: undefined.UndefinedOr[snowflakes.SnowflakeishSequence[PartialUser] | bool] = undefined.UNDEFINED,
        role_mentions: undefined.UndefinedOr[
            snowflakes.SnowflakeishSequence[guilds.PartialRole] | bool
        ] = undefined.UNDEFINED,
        flags: undefined.UndefinedType | int | messages.MessageFlag = undefined.UNDEFINED,
    ) -> typing.NoReturn:
        msg = "Unable to send a DM to yourself"
        raise TypeError(msg)



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import inspect
import sys
import typing

import mock
import pytest

from hikari.internal import reflect


@pytest.mark.skipif(sys.version_info >= (3, 10), reason="This strategy is specific to 3.10 < versions")
class TestResolveSignatureOldStrategy:
    def test_handles_normal_references(self):
        def foo(bar: str, bat: int) -> str: ...

        signature = reflect.resolve_signature(foo)
        assert signature.parameters["bar"].annotation is str
        assert signature.parameters["bat"].annotation is int
        assert signature.return_annotation is str

    def test_handles_normal_no_annotations(self):
        def foo(bar, bat): ...

        signature = reflect.resolve_signature(foo)
        assert signature.parameters["bar"].annotation is reflect.EMPTY
        assert signature.parameters["bat"].annotation is reflect.EMPTY
        assert signature.return_annotation is reflect.EMPTY

    def test_handles_forward_annotated_parameters(self):
        def foo(bar: "str", bat: "int") -> str: ...

        signature = reflect.resolve_signature(foo)
        assert signature.parameters["bar"].annotation is str
        assert signature.parameters["bat"].annotation is int
        assert signature.return_annotation is str

    def test_handles_forward_annotated_return(self):
        def foo(bar: str, bat: int) -> "str": ...

        signature = reflect.resolve_signature(foo)
        assert signature.parameters["bar"].annotation is str
        assert signature.parameters["bat"].annotation is int
        assert signature.return_annotation is str

    def test_handles_forward_annotations(self):
        def foo(bar: "str", bat: "int") -> "str": ...

        signature = reflect.resolve_signature(foo)
        assert signature.parameters["bar"].annotation is str
        assert signature.parameters["bat"].annotation is int
        assert signature.return_annotation is str

    def test_handles_mixed_annotations(self):
        def foo(bar: str, bat: "int"): ...

        signature = reflect.resolve_signature(foo)
        assert signature.parameters["bar"].annotation is str
        assert signature.parameters["bat"].annotation is int
        assert signature.return_annotation is reflect.EMPTY

    def test_handles_None(self):
        def foo(bar: None) -> None: ...

        signature = reflect.resolve_signature(foo)
        assert signature.parameters["bar"].annotation is None
        assert signature.return_annotation is None

    def test_handles_NoneType(self):
        def foo(bar: type(None)) -> type(None): ...

        signature = reflect.resolve_signature(foo)
        assert signature.parameters["bar"].annotation is None
        assert signature.return_annotation is None

    def test_handles_only_return_annotated(self):
        def foo(bar, bat) -> str: ...

        signature = reflect.resolve_signature(foo)
        assert signature.parameters["bar"].annotation is reflect.EMPTY
        assert signature.parameters["bat"].annotation is reflect.EMPTY
        assert signature.return_annotation is str

    def test_handles_nested_annotations(self):
        def foo(bar: typing.Optional[typing.Iterator[int]]): ...

        signature = reflect.resolve_signature(foo)
        assert signature.parameters["bar"].annotation == typing.Optional[typing.Iterator[int]]


@pytest.mark.skipif(sys.version_info < (3, 10), reason="This strategy is specific to 3.10 >= versions")
def test_resolve_signature():
    foo = object()

    with mock.patch.object(inspect, "signature") as signature:
        sig = reflect.resolve_signature(foo)

    assert sig is signature.return_value
    signature.assert_called_once_with(foo, eval_str=True)



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import mock
import pytest

from hikari import files
from hikari.internal import routes
from tests.hikari import hikari_test_helpers


class TestCompiledRoute:
    @pytest.fixture
    def compiled_route(self):
        return routes.CompiledRoute(
            major_param_hash="abc123", route=mock.Mock(method="GET"), compiled_path="/some/endpoint"
        )

    def test_method(self, compiled_route):
        assert compiled_route.method == "GET"

    def test_create_url(self, compiled_route):
        assert compiled_route.create_url("https://some.url/api") == "https://some.url/api/some/endpoint"

    def test_create_real_bucket_hash(self, compiled_route):
        assert compiled_route.create_real_bucket_hash("UNKNOWN", "AUTH_HASH") == "UNKNOWN;AUTH_HASH;abc123"

    def test__str__(self, compiled_route):
        assert str(compiled_route) == "GET /some/endpoint"


class TestRoute:
    @pytest.mark.parametrize(
        ("route", "params"),
        [
            (routes.DELETE_CHANNEL, frozenset(("channel",))),
            (routes.PATCH_GUILD, frozenset(("guild",))),
            (routes.POST_WEBHOOK_WITH_TOKEN, frozenset(("webhook", "token"))),
            (routes.GET_WEBHOOK, frozenset(("webhook",))),
            (routes.GET_INVITE, None),
        ],
    )
    def test_major_params(self, route, params):
        assert route.major_params == params

    def test_compile_with_no_major_params(self):
        route = routes.Route(method="GET", path_template="/some/endpoint/{baguette}")
        expected = routes.CompiledRoute(route=route, compiled_path="/some/endpoint/1234", major_param_hash="-")

        assert route.compile(baguette=1234) == expected

    def test_compile_with_channel_major_params(self):
        route = routes.Route(method="GET", path_template="/channels/{channel}")
        expected = routes.CompiledRoute(route=route, compiled_path="/channels/4325", major_param_hash="4325")

        assert route.compile(channel=4325) == expected

    def test_compile_with_guild_major_params(self):
        route = routes.Route(method="GET", path_template="/guilds/{guild}")
        expected = routes.CompiledRoute(route=route, compiled_path="/guilds/5555", major_param_hash="5555")

        assert route.compile(guild=5555) == expected

    def test_compile_with_webhook_major_params(self):
        route = routes.Route(method="GET", path_template="/webhooks/{webhook}/{token}")
        expected = routes.CompiledRoute(
            route=route, compiled_path="/webhooks/123/okfdkdfkdf", major_param_hash="123:okfdkdfkdf"
        )

        assert route.compile(webhook=123, token="okfdkdfkdf") == expected

    def test__str__(self):
        assert (
            str(routes.Route(method="GET", path_template="/some/endpoint/{channel}")) == "GET /some/endpoint/{channel}"
        )


class TestCDNRoute:
    def test_zero_formats_results_in_error(self):
        with pytest.raises(ValueError, match="/foo/bar must have at least one valid format set"):
            routes.CDNRoute("/foo/bar", set())

    def test_any_formats_results_in_no_error(self):
        routes.CDNRoute("/foo/bar", {"DO", "RAY", "ME"})

    def test_formats_converted_to_frozenset(self):
        route = routes.CDNRoute("/foo/bar", {"I", "REALLY", "LIKE", "CATS"})
        assert isinstance(route.valid_formats, frozenset)
        assert route.valid_formats == {"I", "REALLY", "LIKE", "CATS"}

    def test_formats_converted_to_upper(self):
        route = routes.CDNRoute("/foo/bar", {"FOO", "BaR", "bAz", "bork"})
        assert route.valid_formats == {"FOO", "BAR", "BAZ", "BORK"}

    def test_eq_operator__considers_path_template_only(self):
        route1 = routes.CDNRoute("/foo/bar", {"HELLO", "WORLD"})
        route2 = routes.CDNRoute("/foo/bar", {"I", "SAID", "MEOW"})
        route3 = routes.CDNRoute("/foo/bar", {"I", "SAID", "MEOW"})
        route4 = routes.CDNRoute("/foo/bar/baz", {"I", "SAID", "MEOW"})
        assert route1 == route2
        assert route1 == route3
        assert route1 != route4
        assert route2 == route3
        assert route2 != route4
        assert route3 != route4

    def test_hash_operator_considers_path_template_only(self):
        route1 = routes.CDNRoute("/foo/bar", {"HELLO", "WORLD"})
        route2 = routes.CDNRoute("/foo/bar", {"I", "SAID", "MEOW"})
        route3 = routes.CDNRoute("/foo/bar", {"I", "SAID", "MEOW"})
        route4 = routes.CDNRoute("/foo/bar/baz", {"I", "SAID", "MEOW"})
        assert hash(route1) == hash(route2)
        assert hash(route1) == hash(route3)
        assert hash(route1) != hash(route4)
        assert hash(route2) == hash(route3)
        assert hash(route2) != hash(route4)
        assert hash(route3) != hash(route4)

    @pytest.mark.parametrize(
        ("input_file_format", "expected_file_format"),
        [
            ("PNG", "png"),
            ("JPG", "jpg"),
            ("JPEG", "jpeg"),
            ("WEBP", "webp"),
            ("AWEBP", "webp"),
            ("APNG", "png"),
            ("GIF", "gif"),
            ("LOTTIE", "json"),
        ],
    )
    def test_compile_uses_correct_extensions(self, input_file_format, expected_file_format):
        route = routes.CDNRoute("/foo/bar", {"PNG", "JPG", "JPEG", "WEBP", "AWEBP", "APNG", "GIF", "LOTTIE"})

        compiled_url = route.compile("http://example.com", file_format=input_file_format)

        assert f".{expected_file_format}" in compiled_url, f"compiled_url={compiled_url}"

    def test_size_parameter_when_specified(self):
        route = routes.CDNRoute("/foo/bar", {"APNG", "LOTTIE", "PNG", "JPG", "GIF"})

        compiled_url = route.compile("http://example.com", file_format="PNG", size=128)

        assert f"size=128" in compiled_url, f"compiled_url={compiled_url}"

    def test_compile_includes_lossless_on_webp(self):
        route = routes.CDNRoute("/foo/bar", {"WEBP", "AWEBP", "GIF"})

        compiled_url = route.compile("http://example.com", file_format="WEBP", hash="a_boooob", lossless=False)

        assert "lossless=false" in compiled_url, f"compiled_url={compiled_url}"

    @pytest.mark.parametrize(
        ("input_file_format", "expected_file_format"), [("JPG", "jpg"), ("jpg", "jpg"), ("PNG", "png"), ("PNG", "png")]
    )
    def test_compile_uses_lowercase_file_format_always(self, input_file_format, expected_file_format):
        route = routes.CDNRoute("/foo/bar", {"PNG", "JPG"})

        compiled_url = route.compile("http://example.com", file_format=input_file_format)

        assert compiled_url.endswith(f".{expected_file_format}"), f"compiled_url={compiled_url}"

    def test_disallowed_file_format_raises_TypeError(self):
        route = routes.CDNRoute("/foo/bar", {"PNG", "JPG"})

        with pytest.raises(TypeError, match="not a valid format for this asset"):
            route.compile("http://example.com", file_format="GIF")

    def test_allowed_file_format_does_not_raise_TypeError(self):
        route = routes.CDNRoute("/foo/bar", {"PNG", "JPG"})

        route.compile("http://example.com", file_format="PNG")

    def test_sets_animated_when_requesting_awebp(self):
        route = routes.CDNRoute("/foo/bar", {"AWEBP"})

        compiled_url = route.compile("http://example.com", file_format="AWEBP", hash="a_boooob")

        assert "animated=true" in compiled_url, f"compiled_url={compiled_url}"

    def test_sets_passthrough_when_requesting_png_on_apng_supporting_asset(self):
        route = routes.CDNRoute("/foo/bar", {"PNG", "APNG"})

        compiled_url = route.compile("http://example.com", file_format="PNG", hash="boooob")

        assert "passthrough=false" in compiled_url, f"compiled_url={compiled_url}"

    def test_passing_non_power_of_2_sizes_to_sizable_raises_ValueError(self):
        route = routes.CDNRoute("/foo/bar", {"PNG", "JPG", "GIF"})
        with pytest.raises(ValueError, match="size must be an integer power of 2 between 16 and 4096 inclusive"):
            route.compile("http://example.com", file_format="PNG", hash="boooob", size=123)

    def test_passing_invalid_magnitude_sizes_to_sizable_raises_ValueError(self):
        route = routes.CDNRoute("/foo/bar", {"PNG", "JPG", "GIF"})
        with pytest.raises(ValueError, match="size must be an integer power of 2 between 16 and 4096 inclusive"):
            route.compile("http://example.com", file_format="PNG", hash="boooob", size=int(2**13))

    def test_passing_negative_sizes_to_sizable_raises_ValueError(self):
        route = routes.CDNRoute("/foo/bar", {"PNG", "JPG", "GIF"})
        with pytest.raises(ValueError, match="size must be positive"):
            route.compile("http://example.com", file_format="PNG", hash="boooob", size=-10)

    @pytest.mark.parametrize("size", [int(2**size) for size in range(4, 13)])
    def test_passing_valid_sizes_to_sizable_does_not_raise_ValueError(self, size):
        route = routes.CDNRoute("/foo/bar", {"PNG", "JPG", "GIF"})
        route.compile("http://example.com", file_format="PNG", hash="boooob", size=size)

    def test_passing_size_adds_query_string(self):
        route = routes.CDNRoute("/foo/bar", {"PNG", "JPG", "GIF"})
        compiled_url = route.compile("http://example.com", file_format="PNG", hash="boooob", size=128)
        assert compiled_url.endswith(".png?size=128"), f"compiled_url={compiled_url}"

    def test_passing_no_size_does_not_add_query_string(self):
        route = routes.CDNRoute("/foo/bar", {"PNG", "JPG", "GIF"})
        compiled_url = route.compile("http://example.com", file_format="PNG", hash="boooob")
        assert "size=" not in compiled_url, f"compiled_url={compiled_url}"

    @pytest.mark.parametrize(
        ("base_url", "template", "file_format", "kwds", "foo", "bar", "expected_url"),
        [
            ("http://example.com", "/{foo}/bar", "GIF", {}, "baz", "bork qux", "http://example.com/baz/bar.gif"),
            ("http://example.com", "/{foo}/bar", "GIF", {}, "baz", "bork qux", "http://example.com/baz/bar.gif"),
            (
                "http://example.com",
                "/{foo}/{bar}",
                "PNG",
                {"size": 128},
                "baz",
                "bork qux",
                "http://example.com/baz/bork%20qux.png?size=128&passthrough=false",
            ),
            (
                "http://example.com",
                "/{foo}/{bar}",
                "APNG",
                {},
                "baz",
                "bork qux",
                "http://example.com/baz/bork%20qux.png",
            ),
            (
                "http://example.com",
                "/{foo}/bar",
                "JPG",
                {"size": 128},
                "baz",
                "bork qux",
                "http://example.com/baz/bar.jpg?size=128",
            ),
            (
                "http://example.com",
                "/{foo}/{bar}",
                "WEBP",
                {},
                "baz",
                "bork qux",
                "http://example.com/baz/bork%20qux.webp?lossless=true",
            ),
            (
                "http://example.com",
                "/{foo}/{bar}",
                "WEBP",
                {"lossless": False},
                "baz",
                123456,
                "http://example.com/baz/123456.webp?lossless=false",
            ),
        ],
    )
    def test_compile_generates_expected_url(self, base_url, template, file_format, kwds, foo, bar, expected_url):
        route = routes.CDNRoute(template, {"PNG", "GIF", "JPG", "WEBP", "APNG"})

        actual_url = route.compile(base_url=base_url, file_format=file_format, foo=foo, bar=bar, **kwds)

        assert actual_url == expected_url

    def test_compile_to_file_calls_compile(self):
        route = routes.CDNRoute("/hello/world", {"PNG", "JPG"})

        with mock.patch.object(routes.CDNRoute, "compile") as patched_compile:
            route.compile_to_file(
                "https://blep.com", file_format="FORMAT", size=256, lossless=True, boop="oyy lumo", nya="weeb"
            )

        patched_compile.assert_called_once_with(
            "https://blep.com", file_format="FORMAT", size=256, lossless=True, boop="oyy lumo", nya="weeb"
        )

    def test_compile_to_file_passes_compile_result_to_URL_and_returns_constructed_url(self):
        resultant_url_str = "http://blep.com/hello/world/weeb/oyy%20lumo"
        route = routes.CDNRoute("/hello/world/{nya}/{boop}", {"PNG", "JPG"})

        with mock.patch.object(routes.CDNRoute, "compile", return_value=resultant_url_str):
            result = route.compile_to_file("https://blep.com", file_format="PNG", size=64, boop="oyy lumo", nya="weeb")

        assert result == files.URL(resultant_url_str)



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import mock
import pytest

from hikari import applications
from hikari import components
from hikari import monetization
from hikari import snowflakes
from hikari import traits
from hikari.impl import special_endpoints
from hikari.interactions import base_interactions
from hikari.interactions import modal_interactions


@pytest.fixture
def mock_app():
    return mock.Mock(rest=mock.AsyncMock())


class TestModalInteraction:
    @pytest.fixture
    def mock_modal_interaction(self, mock_app):
        return modal_interactions.ModalInteraction(
            app=mock_app,
            id=snowflakes.Snowflake(2312312),
            type=base_interactions.InteractionType.APPLICATION_COMMAND,
            channel=object(),
            guild_id=snowflakes.Snowflake(5412231),
            member=object(),
            user=object(),
            token="httptptptptptptptp",
            version=1,
            application_id=snowflakes.Snowflake(43123),
            custom_id="OKOKOK",
            message=object(),
            locale="es-ES",
            guild_locale="en-US",
            app_permissions=543123,
            components=special_endpoints.ModalActionRowBuilder(
                id=9817398,
                components=[
                    components.TextInputComponent(
                        type=components.ComponentType.TEXT_INPUT, id=9817398, custom_id="le id", value="le value"
                    )
                ],
            ),
            entitlements=[
                monetization.Entitlement(
                    id=snowflakes.Snowflake(123123),
                    sku_id=snowflakes.Snowflake(123123),
                    application_id=snowflakes.Snowflake(123123),
                    guild_id=snowflakes.Snowflake(123123),
                    user_id=snowflakes.Snowflake(123123),
                    type=monetization.EntitlementType.APPLICATION_SUBSCRIPTION,
                    starts_at=None,
                    ends_at=None,
                    is_deleted=False,
                    subscription_id=None,
                )
            ],
            authorizing_integration_owners={
                applications.ApplicationIntegrationType.GUILD_INSTALL: snowflakes.Snowflake(123)
            },
            context=applications.ApplicationContextType.PRIVATE_CHANNEL,
        )

    def test_build_response(self, mock_modal_interaction, mock_app):
        mock_app.rest.interaction_message_builder = mock.Mock()
        response = mock_modal_interaction.build_response()

        assert response is mock_app.rest.interaction_message_builder.return_value
        mock_app.rest.interaction_message_builder.assert_called_once()

    def test_build_deferred_response(self, mock_modal_interaction, mock_app):
        mock_app.rest.interaction_deferred_builder = mock.Mock()
        response = mock_modal_interaction.build_deferred_response()

        assert response is mock_app.rest.interaction_deferred_builder.return_value
        mock_app.rest.interaction_deferred_builder.assert_called_once()

    @pytest.mark.asyncio
    async def test_fetch_guild(self, mock_modal_interaction, mock_app):
        mock_modal_interaction.guild_id = 43123123

        assert await mock_modal_interaction.fetch_guild() is mock_app.rest.fetch_guild.return_value

        mock_app.rest.fetch_guild.assert_awaited_once_with(43123123)

    @pytest.mark.asyncio
    async def test_fetch_guild_for_dm_interaction(self, mock_modal_interaction, mock_app):
        mock_modal_interaction.guild_id = None

        assert await mock_modal_interaction.fetch_guild() is None

        mock_app.rest.fetch_guild.assert_not_called()

    def test_get_guild(self, mock_modal_interaction, mock_app):
        mock_modal_interaction.guild_id = 874356

        assert mock_modal_interaction.get_guild() is mock_app.cache.get_guild.return_value

        mock_app.cache.get_guild.assert_called_once_with(874356)

    def test_get_guild_for_dm_interaction(self, mock_modal_interaction, mock_app):
        mock_modal_interaction.guild_id = None

        assert mock_modal_interaction.get_guild() is None

        mock_app.cache.get_guild.assert_not_called()

    def test_get_guild_when_cacheless(self, mock_modal_interaction, mock_app):
        mock_modal_interaction.guild_id = 321123
        mock_modal_interaction.app = mock.Mock(traits.RESTAware)

        assert mock_modal_interaction.get_guild() is None

        mock_app.cache.get_guild.assert_not_called()



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Entities directly related to creating and managing gateway shard sessions."""

from __future__ import annotations

__all__: typing.Sequence[str] = ("GatewayBotInfo", "SessionStartLimit")

import typing

import attrs

from hikari.internal import attrs_extensions
from hikari.internal import time

if typing.TYPE_CHECKING:
    import datetime


@attrs_extensions.with_copy
@attrs.define(kw_only=True, weakref_slot=False)
class SessionStartLimit:
    """Used to represent information about the current session start limits."""

    total: int = attrs.field(repr=True)
    """The total number of session starts the current bot is allowed."""

    remaining: int = attrs.field(repr=True)
    """The remaining number of session starts this bot has."""

    reset_after: datetime.timedelta = attrs.field(repr=True)
    """When [`hikari.sessions.SessionStartLimit.remaining`][] will reset for the current bot.

    After it resets it will be set to [`hikari.sessions.SessionStartLimit.total`][].
    """

    # This is not documented at the time of writing, but is a confirmed API
    # feature, so I have included it for completeness.
    max_concurrency: int = attrs.field(repr=True)
    """Maximum connection concurrency.

    This defines how many shards can be started at once within a 5 second
    window. For most bots, this will always be `1`, but for very large bots,
    this may be increased to reduce startup times. Contact Discord for
    more information.
    """

    _created_at: datetime.datetime = attrs.field(factory=time.local_datetime, init=False)

    @property
    def used(self) -> int:
        """Return how many times you have sent an IDENTIFY in the window."""
        return self.total - self.remaining

    @property
    def reset_at(self) -> datetime.datetime:
        """Return the approximate time that the IDENTIFY limit resets at."""
        return self._created_at + self.reset_after


@attrs_extensions.with_copy
@attrs.define(kw_only=True, weakref_slot=False)
class GatewayBotInfo:
    """Used to represent gateway information for the connected bot."""

    url: str = attrs.field(repr=True)
    """The WSS URL that can be used for connecting to the gateway."""

    shard_count: int = attrs.field(repr=True)
    """The recommended number of shards to use when connecting to the gateway."""

    session_start_limit: SessionStartLimit = attrs.field(repr=True)
    """Information about the bot's current session start limit."""



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import timeit
import typing

from hikari.internal import fast_protocol


@typing.runtime_checkable
class BasicPyProtocol(typing.Protocol):
    def test(self, arg1: str, arg2: int) -> list[int]:
        raise NotImplementedError

    def test2(self, arg1: str, arg2: int) -> list[int]:
        raise NotImplementedError

    def test3(self, arg1: str, arg2: int) -> list[int]:
        raise NotImplementedError

    def test4(self, arg1: str, arg2: int) -> list[int]:
        raise NotImplementedError

    def test5(self, arg1: str, arg2: int) -> list[int]:
        raise NotImplementedError

    def test6(self, arg1: str, arg2: int) -> list[int]:
        raise NotImplementedError

    def test7(self, arg1: str, arg2: int) -> list[int]:
        raise NotImplementedError

    def test8(self, arg1: str, arg2: int) -> list[int]:
        raise NotImplementedError

    def test9(self, arg1: str, arg2: int) -> list[int]:
        raise NotImplementedError

    def test10(self, arg1: str, arg2: int) -> list[int]:
        raise NotImplementedError


class BasicHikariProtocol(fast_protocol.FastProtocolChecking, typing.Protocol):
    def test(self, arg1: str, arg2: int) -> list[int]:
        raise NotImplementedError

    def test2(self, arg1: str, arg2: int) -> list[int]:
        raise NotImplementedError

    def test3(self, arg1: str, arg2: int) -> list[int]:
        raise NotImplementedError

    def test4(self, arg1: str, arg2: int) -> list[int]:
        raise NotImplementedError

    def test5(self, arg1: str, arg2: int) -> list[int]:
        raise NotImplementedError

    def test6(self, arg1: str, arg2: int) -> list[int]:
        raise NotImplementedError

    def test7(self, arg1: str, arg2: int) -> list[int]:
        raise NotImplementedError

    def test8(self, arg1: str, arg2: int) -> list[int]:
        raise NotImplementedError

    def test9(self, arg1: str, arg2: int) -> list[int]:
        raise NotImplementedError

    def test10(self, arg1: str, arg2: int) -> list[int]:
        raise NotImplementedError


class Valid:
    def test(self, arg1: str, arg2: int) -> list[int]: ...

    def test2(self, arg1: str, arg2: int) -> list[int]: ...

    def test3(self, arg1: str, arg2: int) -> list[int]: ...

    def test4(self, arg1: str, arg2: int) -> list[int]: ...

    def test5(self, arg1: str, arg2: int) -> list[int]: ...

    def test6(self, arg1: str, arg2: int) -> list[int]: ...

    def test7(self, arg1: str, arg2: int) -> list[int]: ...

    def test8(self, arg1: str, arg2: int) -> list[int]: ...

    def test9(self, arg1: str, arg2: int) -> list[int]: ...

    def test10(self, arg1: str, arg2: int) -> list[int]: ...


class Invalid: ...


isinstance_long = Valid()
isinstance_failfast = Invalid()

# Dummy work to churn the CPU up.
for i in range(100_000):
    assert sum(i for i in range(10)) > 0

py_protocol_isinstance_long_time = timeit.timeit(
    "assert isinstance(isinstance_long, BasicPyProtocol) is True", number=1_000_000, globals=globals()
)
hikari_protocol_isinstance_long_time = timeit.timeit(
    "assert isinstance(isinstance_long, BasicHikariProtocol) is True", number=1_000_000, globals=globals()
)
py_protocol_isinstance_failfast_time = timeit.timeit(
    "assert isinstance(isinstance_failfast, BasicPyProtocol) is False", number=1_000_000, globals=globals()
)
hikari_protocol_isinstance_failfast_time = timeit.timeit(
    "assert isinstance(isinstance_failfast, BasicHikariProtocol) is False", number=1_000_000, globals=globals()
)

py_protocol_issubclass_long_time = timeit.timeit(
    "assert issubclass(Valid, BasicPyProtocol) is True", number=1_000_000, globals=globals()
)
hikari_protocol_issubclass_long_time = timeit.timeit(
    "assert issubclass(Valid, BasicHikariProtocol) is True", number=1_000_000, globals=globals()
)
py_protocol_issubclass_failfast_time = timeit.timeit(
    "assert issubclass(Invalid, BasicPyProtocol) is False", number=1_000_000, globals=globals()
)
hikari_protocol_issubclass_failfast_time = timeit.timeit(
    "assert issubclass(Invalid, BasicHikariProtocol) is False", number=1_000_000, globals=globals()
)

print("isinstance(long, BasicPyProtocol)", py_protocol_isinstance_long_time, "µs")
print("isinstance(long, BasicHikariProtocol)", hikari_protocol_isinstance_long_time, "µs")
print("isinstance(failfast, BasicPyProtocol)", py_protocol_isinstance_failfast_time, "µs")
print("isinstance(failfast, BasicHikariProtocol)", hikari_protocol_isinstance_failfast_time, "µs")

print("issubclass(long, BasicPyProtocol)", py_protocol_issubclass_long_time, "µs")
print("issubclass(long, BasicHikariProtocol)", hikari_protocol_issubclass_long_time, "µs")
print("issubclass(failfast, BasicPyProtocol)", py_protocol_issubclass_failfast_time, "µs")
print("issubclass(failfast, BasicHikariProtocol)", hikari_protocol_issubclass_failfast_time, "µs")



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import pytest

from hikari import emojis
from hikari import snowflakes


class TestEmoji:
    @pytest.mark.parametrize(
        ("input", "output"),
        [
            ("<:foo:12345>", emojis.CustomEmoji(id=snowflakes.Snowflake(12345), name="foo", is_animated=False)),
            ("<bar:foo:12345>", emojis.CustomEmoji(id=snowflakes.Snowflake(12345), name="foo", is_animated=False)),
            ("<a:foo:12345>", emojis.CustomEmoji(id=snowflakes.Snowflake(12345), name="foo", is_animated=True)),
            ("\N{OK HAND SIGN}", emojis.UnicodeEmoji("\N{OK HAND SIGN}")),
            (
                "\N{REGIONAL INDICATOR SYMBOL LETTER G}\N{REGIONAL INDICATOR SYMBOL LETTER B}",
                emojis.UnicodeEmoji("\N{REGIONAL INDICATOR SYMBOL LETTER G}\N{REGIONAL INDICATOR SYMBOL LETTER B}"),
            ),
        ],
    )
    def test_parse(self, input, output):
        assert emojis.Emoji.parse(input) == output


class TestUnicodeEmoji:
    @pytest.fixture
    def emoji(self):
        return emojis.UnicodeEmoji("\N{OK HAND SIGN}")

    def test_name_property(self, emoji):
        assert emoji.name == emoji

    def test_url_name_property(self, emoji):
        assert emoji.url_name == emoji

    def test_mention_property(self, emoji):
        assert emoji.mention == emoji

    def test_codepoints_property(self, emoji):
        assert emoji.codepoints == [128076]

    @pytest.mark.parametrize(
        ("codepoints", "expected_filename"),
        [
            # Normal tests
            ([0x1F44C], "1f44c"),
            ([0x1F44C, 0x1F44C, 0x1F44C, 0x1F44C], "1f44c-1f44c-1f44c-1f44c"),
            # Outliers tests
            # 1. Second codepoint is not 0xFE0F => Nothing
            ([0xFE0F, 0x1F44C], "fe0f-1f44c"),
            # 2. More than 4 codepoints => Nothing
            ([0x1F44C, 0xFE0F, 0x1F44C, 0x1F44C, 0x1F44C], "1f44c-fe0f-1f44c-1f44c-1f44c"),
            # 3. Third codepoint is 0x200D => Nothing
            ([0x1F44C, 0xFE0F, 0x200D], "1f44c-fe0f-200d"),
            # 4. None of above apply => Remove second codepoint
            ([0x200D, 0xFE0F, 0x1F44C, 0x1F44C], "200d-1f44c-1f44c"),
        ],
    )
    def test_filename_property(self, codepoints, expected_filename):
        emoji = emojis.UnicodeEmoji.parse_codepoints(*codepoints)
        assert emoji.filename == f"{expected_filename}.png"

    def test_url_property(self, emoji):
        assert emoji.url == "https://raw.githubusercontent.com/discord/twemoji/master/assets/72x72/1f44c.png"

    def test_unicode_escape_property(self, emoji):
        assert emoji.unicode_escape == "\\U0001f44c"

    def test_parse_codepoints(self, emoji):
        assert emojis.UnicodeEmoji.parse_codepoints(128076) == emoji

    def test_parse_unicode_escape(self, emoji):
        assert emojis.UnicodeEmoji.parse_unicode_escape("\\U0001f44c") == emoji

    def test_str_operator(self, emoji):
        assert str(emoji) == emoji

    @pytest.mark.parametrize(
        ("input", "output"),
        [
            ("\N{OK HAND SIGN}", emojis.UnicodeEmoji("\N{OK HAND SIGN}")),
            (
                "\N{REGIONAL INDICATOR SYMBOL LETTER G}\N{REGIONAL INDICATOR SYMBOL LETTER B}",
                emojis.UnicodeEmoji("\N{REGIONAL INDICATOR SYMBOL LETTER G}\N{REGIONAL INDICATOR SYMBOL LETTER B}"),
            ),
        ],
    )
    def test_parse(self, input, output):
        assert emojis.UnicodeEmoji.parse(input) == output


class TestCustomEmoji:
    @pytest.fixture
    def emoji(self):
        return emojis.CustomEmoji(id=3213452, name="ok", is_animated=False)

    def test_filename_property(self, emoji):
        assert emoji.filename == "3213452.png"

    def test_filename_property_when_animated(self, emoji):
        emoji.is_animated = True
        assert emoji.filename == "3213452.gif"

    def test_url_name_property(self, emoji):
        assert emoji.url_name == "ok:3213452"

    def test_mention_property(self, emoji):
        assert emoji.mention == "<:ok:3213452>"

    def test_mention_property_when_animated(self, emoji):
        emoji.is_animated = True

        assert emoji.mention == "<a:ok:3213452>"

    def test_url_property(self, emoji):
        assert emoji.url == "https://cdn.discordapp.com/emojis/3213452.png"

    def test_str_operator_when_populated_name(self):
        emoji = emojis.CustomEmoji(id=snowflakes.Snowflake(12345), name="peepoSad", is_animated=True)
        assert str(emoji) == emoji.mention

    @pytest.mark.parametrize(
        ("input", "output"),
        [
            ("<:foo:12345>", emojis.CustomEmoji(id=snowflakes.Snowflake(12345), name="foo", is_animated=False)),
            ("<bar:foo:12345>", emojis.CustomEmoji(id=snowflakes.Snowflake(12345), name="foo", is_animated=False)),
            ("<a:foo:12345>", emojis.CustomEmoji(id=snowflakes.Snowflake(12345), name="foo", is_animated=True)),
        ],
    )
    def test_parse(self, input, output):
        assert emojis.CustomEmoji.parse(input) == output

    def test_parse_unhappy_path(self):
        with pytest.raises(ValueError, match="Expected an emoji mention"):
            emojis.CustomEmoji.parse("xxx")



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import pytest

from hikari import iterators
from tests.hikari import hikari_test_helpers


class TestLazyIterator:
    @pytest.fixture
    def lazy_iterator(self):
        return hikari_test_helpers.mock_class_namespace(iterators.LazyIterator)()

    def test_asynchronous_only(self, lazy_iterator):
        with pytest.raises(TypeError, match="is async-only, did you mean 'async for' or `anext`?"):
            next(lazy_iterator)

    @pytest.mark.asyncio
    async def test_flatten(self):
        iterator = iterators.FlatLazyIterator([[123, 321, 4352, 123], [], [12343123, 4234432], [543123123]])

        assert await iterator.flatten() == [123, 321, 4352, 123, 12343123, 4234432, 543123123]



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Entities that are used to describe auto-moderation on Discord."""

from __future__ import annotations

__all__: typing.Sequence[str] = (
    "AutoModActionType",
    "AutoModBlockMessage",
    "AutoModEventType",
    "AutoModKeywordPresetType",
    "AutoModRule",
    "AutoModSendAlertMessage",
    "AutoModTimeout",
    "AutoModTriggerType",
    "KeywordPresetTrigger",
    "KeywordTrigger",
    "MemberProfileTrigger",
    "MentionSpamTrigger",
    "PartialAutoModAction",
    "PartialAutoModTrigger",
    "SpamTrigger",
)

import typing

import attrs

from hikari import snowflakes
from hikari.internal import attrs_extensions
from hikari.internal import enums

if typing.TYPE_CHECKING:
    import datetime

    from hikari import traits


class AutoModActionType(int, enums.Enum):
    """The type of an auto-moderation rule action."""

    BLOCK_MESSAGE = 1
    """Block the content of the triggering message."""

    SEND_ALERT_MESSAGE = 2
    """Log the triggering user content to a specified channel."""

    TIMEOUT = 3
    """Timeout the triggering message's author for a specified duration.

    This type can only be set for `KEYWORD` and `MENTION_SPAM` rules.

    Requires the `MODERATE_MEMBERS` permission to use.
    """

    BLOCK_MEMBER_INTERACTION = 4
    """Prevents a member from using text, voice, or other interactions."""


class AutoModEventType(int, enums.Enum):
    """Type of event to check for an auto-moderation rule."""

    MESSAGE_SEND = 1
    """When a member sends or edits a message in the guild."""

    MEMBER_UPDATE = 2
    """When a member updates their guild or user profile."""


class AutoModTriggerType(int, enums.Enum):
    """Type of trigger for an auto-moderation rule."""

    KEYWORD = 1
    """Match message content against a list of keywords and regexes."""

    SPAM = 3
    """Discord's guild anti-spam system."""

    KEYWORD_PRESET = 4
    """Discord's preset keyword triggers."""

    MENTION_SPAM = 5
    """Match messages that exceed the allowed limit of role or user mentions."""

    MEMBER_PROFILE = 6
    """Match user profiles against a list of keywords and regexes."""


class AutoModKeywordPresetType(int, enums.Enum):
    """Discord's KEYWORD_PRESET type."""

    PROFANITY = 1
    """Trigger on words which may be considered forms of swearing or cursing."""

    SEXUAL_CONTENT = 2
    """Trigger on words that refer to explicit behaviour or activity."""

    SLURS = 3
    """Trigger on personal insults and words which "may be considered hate speech"."""


@attrs_extensions.with_copy
@attrs.define(kw_only=True, weakref_slot=False)
class PartialAutoModAction:
    """Base class for an action which is executed when a rule is triggered."""

    type: AutoModActionType = attrs.field()
    """The type of auto-moderation action."""


@attrs.define(kw_only=True, weakref_slot=False)
class AutoModBlockMessage(PartialAutoModAction):
    """Block the content of the triggering message."""


@attrs.define(kw_only=True, weakref_slot=False)
class AutoModSendAlertMessage(PartialAutoModAction):
    """Log the triggering content to a specific channel."""

    channel_id: snowflakes.Snowflake = attrs.field()
    """ID of the channel to log trigger events to."""


@attrs.define(kw_only=True, weakref_slot=False)
class AutoModTimeout(PartialAutoModAction):
    """Timeout the triggering message's author for a specified duration.

    This type can only be set for `KEYWORD` and `MENTION_SPAM` rules.
    """

    duration: datetime.timedelta = attrs.field()
    """The total seconds to timeout the user for (max 2419200 seconds/4 weeks)."""


@attrs.define(kw_only=True, weakref_slot=False)
class AutoModBlockMemberAction(PartialAutoModAction):
    """Prevents a member from using text, voice, or other interactions."""


@attrs_extensions.with_copy
@attrs.define(kw_only=True, weakref_slot=False)
class PartialAutoModTrigger:
    """Base class representing the content a rule triggers on."""

    type: AutoModTriggerType = attrs.field(eq=False, hash=False, repr=False)
    """The type action this triggers."""


@attrs.define(kw_only=True, weakref_slot=False)
class KeywordTrigger(PartialAutoModTrigger):
    """A trigger based on matching message content against a list of keywords."""

    keyword_filter: typing.Sequence[str] = attrs.field(eq=False, hash=False, repr=False)
    """The filter strings this trigger checks for."""

    regex_patterns: typing.Sequence[str] = attrs.field(eq=False, hash=False, repr=False)
    """The filter regexes this trigger checks for."""

    allow_list: typing.Sequence[str] = attrs.field(eq=False, hash=False, repr=False)
    """A sequence of filters which will be exempt from triggering the preset trigger."""


@attrs.define(kw_only=True, weakref_slot=False)
class SpamTrigger(PartialAutoModTrigger):
    """A trigger based on Discord's spam detection."""


@attrs.define(kw_only=True, weakref_slot=False)
class KeywordPresetTrigger(PartialAutoModTrigger):
    """A trigger based on a predefined set of presets provided by Discord."""

    allow_list: typing.Sequence[str] = attrs.field(eq=False, hash=False, repr=False)
    """A sequence of filters which will be exempt from triggering the preset trigger."""

    presets: typing.Sequence[AutoModKeywordPresetType | int] = attrs.field(eq=False, hash=False, repr=False)
    """The predefined presets provided by Discord to match against."""


@attrs.define(kw_only=True, weakref_slot=False)
class MentionSpamTrigger(PartialAutoModTrigger):
    """A trigger based on matching mention spams in message content."""

    mention_total_limit: int = attrs.field(eq=False, hash=False, repr=False)
    """Total number of unique role and user mentions allowed per message."""

    mention_raid_protection_enabled: bool = attrs.field(eq=False, hash=False, repr=False)
    """Whether to automatically detect mention raids."""


@attrs.define(kw_only=True, weakref_slot=False)
class MemberProfileTrigger(PartialAutoModTrigger):
    """A trigger based on matching user profile content against a list of keywords."""

    keyword_filter: typing.Sequence[str] = attrs.field(eq=False, hash=False, repr=False)
    """The filter strings this trigger checks for.

    This supports a wildcard matching strategy which is documented at
    <https://discord.com/developers/docs/resources/auto-moderation#auto-moderation-rule-object-keyword-matching-strategies>.
    """

    regex_patterns: typing.Sequence[str] = attrs.field(eq=False, hash=False, repr=False)
    """The filter regexes this trigger checks for."""

    allow_list: typing.Sequence[str] = attrs.field(eq=False, hash=False, repr=False)
    """A sequence of filters which will be exempt from triggering the preset trigger."""


@attrs_extensions.with_copy
@attrs.define(hash=True, kw_only=True, weakref_slot=False)
class AutoModRule(snowflakes.Unique):
    """Auto moderation rule which defines how user content is filtered."""

    app: traits.RESTAware = attrs.field(
        repr=False, eq=False, hash=False, metadata={attrs_extensions.SKIP_DEEP_COPY: True}
    )
    """The client application that models may use for procedures."""

    id: snowflakes.Snowflake = attrs.field(eq=True, hash=True, repr=True)
    """The ID of this entity."""

    guild_id: snowflakes.Snowflake = attrs.field(eq=False, hash=False, repr=True)
    """ID of the guild this rule belongs to."""

    name: str = attrs.field(eq=False, hash=False, repr=True)
    """The rule's name."""

    creator_id: snowflakes.Snowflake = attrs.field(eq=False, hash=False, repr=True)
    """ID of the user who originally created this rule."""

    event_type: AutoModEventType = attrs.field(eq=False, hash=False, repr=True)
    """The type of event this rule triggers on."""

    trigger: PartialAutoModTrigger = attrs.field(eq=False, hash=False, repr=False)
    """The content this rule triggers on."""

    actions: typing.Sequence[PartialAutoModAction] = attrs.field(eq=False, hash=False, repr=False)
    """Sequence of the actions which will execute when this rule is triggered."""

    is_enabled: bool = attrs.field(eq=False, hash=False, repr=False)
    """Whether this rule is enabled."""

    exempt_role_ids: typing.Sequence[snowflakes.Snowflake] = attrs.field(eq=False, hash=False, repr=False)
    """A sequence of IDs of (up to 50) roles which aren't effected by this rule."""

    exempt_channel_ids: typing.Sequence[snowflakes.Snowflake] = attrs.field(eq=False, hash=False, repr=False)
    """A sequence of IDs of (up to 20) channels which aren't effected by this rule."""



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import attrs
import mock
import pytest

from hikari import intents
from hikari.api import shard as gateway_shard
from hikari.events import base_events


@base_events.requires_intents(intents.Intents.GUILDS)
@attrs.define(eq=False, init=False, kw_only=True)
class DummyGuildEvent(base_events.Event):
    pass


@base_events.no_recursive_throw()
@base_events.requires_intents(intents.Intents.GUILD_PRESENCES)
@attrs.define(eq=False, init=False, kw_only=True)
class DummyPresenceEvent(base_events.Event):
    pass


@base_events.no_recursive_throw()
@attrs.define(eq=False, init=False, kw_only=True)
class ErrorEvent(base_events.Event):
    pass


@attrs.define(eq=False, init=False, kw_only=True)
class DummyGuildDerivedEvent(DummyGuildEvent):
    pass


@attrs.define(eq=False, init=False, kw_only=True)
class DummyPresenceDerivedEvent(DummyPresenceEvent):
    pass


def test_is_no_recursive_throw_event_marked():
    assert base_events.is_no_recursive_throw_event(DummyPresenceEvent)
    assert base_events.is_no_recursive_throw_event(ErrorEvent)
    assert not base_events.is_no_recursive_throw_event(DummyGuildEvent)
    assert not base_events.is_no_recursive_throw_event(DummyGuildDerivedEvent)


def test_requires_intents():
    assert list(base_events.get_required_intents_for(DummyGuildEvent)) == [intents.Intents.GUILDS]
    assert list(base_events.get_required_intents_for(DummyPresenceEvent)) == [intents.Intents.GUILD_PRESENCES]
    assert list(base_events.get_required_intents_for(ErrorEvent)) == []


def test_inherited_requires_intents():
    assert list(base_events.get_required_intents_for(DummyPresenceDerivedEvent)) == [intents.Intents.GUILD_PRESENCES]
    assert list(base_events.get_required_intents_for(DummyGuildDerivedEvent)) == [intents.Intents.GUILDS]


def test_inherited_is_no_recursive_throw_event():
    assert base_events.is_no_recursive_throw_event(DummyPresenceDerivedEvent)
    assert not base_events.is_no_recursive_throw_event(DummyGuildDerivedEvent)


class TestExceptionEvent:
    @pytest.fixture(scope="class")  # we don't modify this so make it once.
    def error(self):
        # Raise and catch to fill in the traceback attribute.
        try:
            raise RuntimeError("blah")
        except RuntimeError as ex:
            return ex

    @pytest.fixture
    def event(self, error):
        return base_events.ExceptionEvent(
            exception=error, failed_event=mock.Mock(base_events.Event), failed_callback=mock.AsyncMock()
        )

    def test_app_property(self, event):
        app = mock.Mock()
        event.failed_event.app = app
        assert event.app is app

    @pytest.mark.parametrize("has_shard", [True, False])
    def test_shard_property(self, has_shard, event):
        shard = mock.Mock(spec_set=gateway_shard.GatewayShard)
        if has_shard:
            event.failed_event.shard = shard
            assert event.shard is shard
        else:
            assert event.shard is None

    def test_exc_info_property(self, event, error):
        assert event.exc_info == (type(error), error, error.__traceback__)

    @pytest.mark.asyncio
    async def test_retry(self, event):
        await event.retry()
        event.failed_callback.assert_awaited_once_with(event.failed_event)



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Application and entities that are used to describe guild templates on Discord."""

from __future__ import annotations

__all__: typing.Sequence[str] = ("Template", "TemplateGuild", "TemplateRole")

import typing

import attrs

from hikari import guilds
from hikari import undefined
from hikari.internal import attrs_extensions
from hikari.internal import typing_extensions

if typing.TYPE_CHECKING:
    import datetime

    from hikari import channels as channels_
    from hikari import colors
    from hikari import permissions as permissions_
    from hikari import snowflakes
    from hikari import traits
    from hikari import users


@attrs_extensions.with_copy
@attrs.define(unsafe_hash=True, kw_only=True, weakref_slot=False)
class TemplateRole(guilds.PartialRole):
    """The partial role object attached to [`hikari.templates.Template`][]."""

    permissions: permissions_.Permissions = attrs.field(eq=False, hash=False, repr=False)
    """The guild wide permissions this role gives to the members it's attached to.

    This may be overridden by channel overwrites.
    """

    color: colors.Color = attrs.field(eq=False, hash=False, repr=True)
    """The colour of this role.

    This will be applied to a member's name in chat if it's their top coloured role.
    """

    is_hoisted: bool = attrs.field(eq=False, hash=False, repr=True)
    """Whether this role is hoisting the members it's attached to in the member list.

    members will be hoisted under their highest role where this is set to [`True`][].
    """

    is_mentionable: bool = attrs.field(eq=False, hash=False, repr=False)
    """Whether this role can be mentioned by all regardless of permissions."""


@attrs_extensions.with_copy
@attrs.define(unsafe_hash=True, kw_only=True, weakref_slot=False)
class TemplateGuild(guilds.PartialGuild):
    """The partial guild object attached to [`hikari.templates.Template`][]."""

    description: str | None = attrs.field(eq=False, hash=False, repr=False)
    """The guild's description, if set."""

    verification_level: guilds.GuildVerificationLevel | int = attrs.field(eq=False, hash=False, repr=False)
    """The verification level needed for a user to participate in this guild."""

    default_message_notifications: guilds.GuildMessageNotificationsLevel | int = attrs.field(
        eq=False, hash=False, repr=False
    )
    """The default setting for message notifications in this guild."""

    explicit_content_filter: guilds.GuildExplicitContentFilterLevel | int = attrs.field(
        eq=False, hash=False, repr=False
    )
    """The setting for the explicit content filter in this guild."""

    preferred_locale: str = attrs.field(eq=False, hash=False, repr=False)
    """The preferred locale to use for this guild.

    This can only be change if [`hikari.guilds.GuildFeature.COMMUNITY`][] is in [`hikari.guilds.Guild.features`][]
    for this guild and will otherwise default to `en-US`.
    """

    afk_timeout: datetime.timedelta = attrs.field(eq=False, hash=False, repr=False)
    """Timeout for activity before a member is classed as AFK.

    How long a voice user has to be AFK for before they are classed as being
    AFK and are moved to the AFK channel ([`hikari.guilds.Guild.afk_channel_id`][]).
    """

    roles: typing.Mapping[snowflakes.Snowflake, TemplateRole] = attrs.field(eq=False, hash=False, repr=False)
    """The roles in the guild.

    !!! note
        [`hikari.guilds.Role.id`][] will be a unique placeholder on all the role
        objects found attached this template guild.
    """

    channels: typing.Mapping[snowflakes.Snowflake, channels_.GuildChannel] = attrs.field(
        eq=False, hash=False, repr=False
    )
    """The channels for the guild.

    !!! note
        [`hikari.channels.GuildChannel.id`][] will be a unique placeholder on all
        the channel objects found attached this template guild.
    """

    afk_channel_id: snowflakes.Snowflake | None = attrs.field(eq=False, hash=False, repr=False)
    """The ID for the channel that AFK voice users get sent to.

    If [`None`][], then no AFK channel is set up for this guild.
    """

    system_channel_id: snowflakes.Snowflake | None = attrs.field(eq=False, hash=False, repr=False)
    """The ID of the system channel or [`None`][] if it is not enabled.

    Welcome messages and Nitro boost messages may be sent to this channel.
    """

    system_channel_flags: guilds.GuildSystemChannelFlag = attrs.field(eq=False, hash=False, repr=False)
    """Return flags for the guild system channel.

    These are used to describe which notifications are suppressed.
    """


@attrs_extensions.with_copy
@attrs.define(unsafe_hash=True, kw_only=True, weakref_slot=False)
class Template:
    """Represents a template used for creating guilds."""

    app: traits.RESTAware = attrs.field(
        repr=False, eq=False, hash=False, metadata={attrs_extensions.SKIP_DEEP_COPY: True}
    )
    """Client application that models may use for procedures."""

    code: str = attrs.field(hash=True, repr=True)
    """The template's unique ID."""

    name: str = attrs.field(eq=False, hash=False, repr=True)
    """The template's name."""

    description: str | None = attrs.field(eq=False, hash=False, repr=False)
    """The template's description."""

    usage_count: int = attrs.field(eq=False, hash=False, repr=True)
    """The number of times the template has been used to create a guild."""

    creator: users.User = attrs.field(eq=False, hash=False, repr=False)
    """The user who created the template."""

    created_at: datetime.datetime = attrs.field(eq=False, hash=False, repr=True)
    """When the template was created."""

    updated_at: datetime.datetime = attrs.field(eq=False, hash=False, repr=True)
    """When the template was last synced with the source guild."""

    source_guild: TemplateGuild = attrs.field(eq=False, hash=False, repr=True)
    """The partial object of the guild this template is based on."""

    is_unsynced: bool = attrs.field(eq=False, hash=False, repr=False)
    """Whether this template is missing changes from it's source guild."""

    async def fetch_self(self) -> Template:
        """Fetch an up-to-date view of this template from the API.

        Returns
        -------
        hikari.templates.Template
            An up-to-date view of this template.

        Raises
        ------
        hikari.errors.UnauthorizedError
            If you are unauthorized to make the request (invalid/missing token).
        hikari.errors.NotFoundError
            If the template is not found.
        hikari.errors.RateLimitTooLongError
            Raised in the event that a rate limit occurs that is
            longer than `max_rate_limit` when making a request.
        hikari.errors.InternalServerError
            If an internal error occurs on Discord while handling the request.
        """
        return await self.app.rest.fetch_template(self.code)

    async def edit(
        self,
        *,
        name: undefined.UndefinedOr[str] = undefined.UNDEFINED,
        description: undefined.UndefinedNoneOr[str] = undefined.UNDEFINED,
    ) -> Template:
        """Modify a guild template.

        Parameters
        ----------
        name
            The name to set for this template.
        description
            The description to set for the template.

        Returns
        -------
        hikari.templates.Template
            The object of the edited template.

        Raises
        ------
        hikari.errors.ForbiddenError
            If you are not part of the guild.
        hikari.errors.NotFoundError
            If the guild is not found or you are missing the [`hikari.permissions.Permissions.MANAGE_GUILD`][]
            permission.
        hikari.errors.UnauthorizedError
            If you are unauthorized to make the request (invalid/missing token).
        hikari.errors.RateLimitTooLongError
            Raised in the event that a rate limit occurs that is longer than max_rate_limit when making a request.
        hikari.errors.InternalServerError
            If an internal error occurs on Discord while handling the request.
        """
        return await self.app.rest.edit_template(self.source_guild, self, name=name, description=description)

    async def delete(self) -> None:
        """Delete a guild template.

        Raises
        ------
        hikari.errors.ForbiddenError
            If you are not part of the guild.
        hikari.errors.NotFoundError
            If the guild is not found or you are missing the [`hikari.permissions.Permissions.MANAGE_GUILD`][]
            permission.
        hikari.errors.UnauthorizedError
            If you are unauthorized to make the request (invalid/missing token).
        hikari.errors.RateLimitTooLongError
            Raised in the event that a rate limit occurs that is longer than max_rate_limit when making a request.
        hikari.errors.InternalServerError
            If an internal error occurs on Discord while handling the request.
        """
        await self.app.rest.delete_template(self.source_guild, self)

    async def sync(self) -> Template:
        """Sync a guild template.

        Returns
        -------
        hikari.templates.Template
            The object of the synced template.

        Raises
        ------
        hikari.errors.ForbiddenError
            If you are not part of the guild or are missing the [`hikari.permissions.Permissions.MANAGE_GUILD`][]
            permission.
        hikari.errors.NotFoundError
            If the guild or template is not found.
        hikari.errors.UnauthorizedError
            If you are unauthorized to make the request (invalid/missing token).
        hikari.errors.RateLimitTooLongError
            Raised in the event that a rate limit occurs that is longer than max_rate_limit when making a request.
        hikari.errors.InternalServerError
            If an internal error occurs on Discord while handling the request.
        """
        return await self.app.rest.sync_guild_template(self.source_guild.id, self.code)

    @typing_extensions.override
    def __str__(self) -> str:
        return f"https://discord.new/{self.code}"



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import math

import pytest

from hikari import colors

tuple_str_happy_test_data = [
    # No brackets, no commas
    ("1 2           3", colors.Color(0x010203)),
    ("0 0 0", colors.Color(0x0)),
    ("0.0 0.0 0.0", colors.Color(0x0)),
    ("0.5 0.6 0.7", colors.Color(0x7F99B2)),
    ("127 153 178", colors.Color(0x7F99B2)),
    ("255 255 255", colors.Color(0xFFFFFF)),
    ("1.0 1.0 1.0", colors.Color(0xFFFFFF)),
    # Parenthesis, no commas
    ("(1 2      3)", colors.Color(0x010203)),
    ("(0 0 0)", colors.Color(0x0)),
    ("(0.0 0.0 0.0)", colors.Color(0x0)),
    ("(0.5 0.6 0.7)", colors.Color(0x7F99B2)),
    ("(127 153 178)", colors.Color(0x7F99B2)),
    ("(255 255 255)", colors.Color(0xFFFFFF)),
    ("(1.0 1.0 1.0)", colors.Color(0xFFFFFF)),
    # Square brackets, no commas
    ("[1 2 3]", colors.Color(0x010203)),
    ("[0 0 0]", colors.Color(0x0)),
    ("[0.0 0.0 0.0]", colors.Color(0x0)),
    ("[0.5 0.6 0.7]", colors.Color(0x7F99B2)),
    ("[127 153 178]", colors.Color(0x7F99B2)),
    ("[255 255 255]", colors.Color(0xFFFFFF)),
    ("[1.0 1.0 1.0]", colors.Color(0xFFFFFF)),
    # Braces, no commas
    ("{1 2 3}", colors.Color(0x010203)),
    ("{0 0 0}", colors.Color(0x0)),
    ("{0.0 0.0 0.0}", colors.Color(0x0)),
    ("{0.5 0.6 0.7}", colors.Color(0x7F99B2)),
    ("{127 153 178}", colors.Color(0x7F99B2)),
    ("{255 255 255}", colors.Color(0xFFFFFF)),
    ("{1.0 1.0 1.0}", colors.Color(0xFFFFFF)),
    # Angle brackets, no commas
    ("< 1 2   3>", colors.Color(0x010203)),
    ("<0 0  0>", colors.Color(0x0)),
    ("<0.0 0.0 0.0>", colors.Color(0x0)),
    ("<0.5\n0.6\n\t0.7>", colors.Color(0x7F99B2)),
    ("<127 153 178>", colors.Color(0x7F99B2)),
    ("<255 255 255>", colors.Color(0xFFFFFF)),
    ("<1.0 1.0 1.0>", colors.Color(0xFFFFFF)),
    # No brackets, commas
    ("1, 2  , 3", colors.Color(0x010203)),
    ("0, 0, 0", colors.Color(0x0)),
    ("0.0, 0.0, 0.0", colors.Color(0x0)),
    ("0.5, 0.6, 0.7", colors.Color(0x7F99B2)),
    ("127, 153, 178", colors.Color(0x7F99B2)),
    ("255, 255, 255", colors.Color(0xFFFFFF)),
    ("1.0, 1.0, 1.0", colors.Color(0xFFFFFF)),
    # Parenthesis, commas
    ("(1, 2  , 3)", colors.Color(0x010203)),
    ("(0,0,0)", colors.Color(0x0)),
    ("(0.0, 0.0, 0.0)", colors.Color(0x0)),
    ("(0.5, 0.6, 0.7)", colors.Color(0x7F99B2)),
    ("(127, 153, 178)", colors.Color(0x7F99B2)),
    ("(255, 255, 255)", colors.Color(0xFFFFFF)),
    ("(1.0, 1.0, 1.0)", colors.Color(0xFFFFFF)),
    # Square brackets, commas
    ("[1, 2, 3]", colors.Color(0x010203)),
    ("[0,0,0]", colors.Color(0x0)),
    ("[0., .0, .0]", colors.Color(0x0)),
    ("[.5, .6, .7]", colors.Color(0x7F99B2)),
    ("[127, 153, 178]", colors.Color(0x7F99B2)),
    ("[255, 255, 255]", colors.Color(0xFFFFFF)),
    ("[1.0, 1.0, 1.0]", colors.Color(0xFFFFFF)),
    # Braces, commas
    ("{1, 2, 3}", colors.Color(0x010203)),
    ("{0, 0, 0}", colors.Color(0x0)),
    ("{0.0, 0.0, 0.0}", colors.Color(0x0)),
    ("{0.5, 0.6, 0.7}", colors.Color(0x7F99B2)),
    ("{127, 153, 178}", colors.Color(0x7F99B2)),
    ("{255, 255, 255}", colors.Color(0xFFFFFF)),
    ("{1.0, 1.0, 1.0}", colors.Color(0xFFFFFF)),
    # Angle brackets, commas
    ("<1, 2, 3>", colors.Color(0x010203)),
    ("<0, 0, 0>", colors.Color(0x0)),
    ("<0.0, 0.0, 0.0>", colors.Color(0x0)),
    ("<0.5, 0.6, 0.7>", colors.Color(0x7F99B2)),
    ("<127, 153, 178>", colors.Color(0x7F99B2)),
    ("<255, 255, 255>", colors.Color(0xFFFFFF)),
    ("<1.0, 1.0, 1.0>", colors.Color(0xFFFFFF)),
]

tuple_str_sad_test_data = [
    # Empty inputs
    ("{}", r"Expected three comma/space separated values"),
    ("[]", r"Expected three comma/space separated values"),
    ("()", r"Expected three comma/space separated values"),
    ("<>", r"Expected three comma/space separated values"),
    # Letters in inputs
    ("{a, 1, 2}", r"Expected digits only for red channel"),
    ("[1, b, 2]", r"Expected digits only for green channel"),
    ("(1, 2, c)", r"Expected digits only for blue channel"),
    ("<d, d, d>", r"Expected digits only for red channel"),
    ("e, e, e", r"Expected digits only for red channel"),
    ("{a 1 2}", r"Expected digits only for red channel"),
    ("[1 b 2]", r"Expected digits only for green channel"),
    ("(1 2 c)", r"Expected digits only for blue channel"),
    ("<d d d>", r"Expected digits only for red channel"),
    ("e e e", r"Expected digits only for red channel"),
    # Mixing ints and floats
    ("1.0, 1.0, 3", r'Expected exactly 1 decimal point "\." in blue channel'),
    ("1.0 2 3", r'Expected exactly 1 decimal point "\." in green channel'),
    # Weird decimals out of range
    ("0.1 2. 0.5", r"Expected green channel to be a decimal in the inclusive range of 0.0 and 1.0"),
    # Too many int digits
    ("0 25600 200", r"Expected 1 to 3 digits for green channel, got 5"),
    # Ints out of range
    ("0 256 200", r"Expected green channel to be in the inclusive range of 0 and 255, got '256'"),
    # Floats out of range
    ("0.0 1.1 0.5", r"Expected green channel to be in the inclusive range of 0.0 and 1.0"),
    # Too few ints
    ("10 20", r"Expected three comma/space separated values"),
    # Too many ints
    ("10 20 30 40", r"Expected three comma/space separated values"),
    # Too few floats
    ("0.5", r"Expected three comma/space separated values"),
    ("0.5 0.5", r"Expected three comma/space separated values"),
    # Too many floats
    ("0.5 0.5 0.5 0.5", r"Expected three comma/space separated values"),
    # Too many commas
    ("100,,100, 100", r"Expected three comma/space separated values"),
    # Mixing commas and spaces without commas
    ("100,100 100", r"Expected three comma/space separated values"),
]


class TestColor:
    @pytest.mark.parametrize("i", [0, 0x1, 0x11, 0x111, 0x1111, 0x11111, 0xFFFFFF])
    def test_Color_validates_constructor_and_passes_for_valid_values(self, i):
        assert colors.Color(i) is not None

    @pytest.mark.parametrize("i", [-1, 0x1000000])
    def test_Color_validates_constructor_and_fails_for_out_of_range_values(self, i):
        with pytest.raises(ValueError, match=r"raw_rgb must be in the exclusive range of 0 and 16777215"):
            colors.Color(i)

    @pytest.mark.parametrize("i", [0, 0x1, 0x11, 0x111, 0x1111, 0x11111, 0xFFFFFF])
    def test_Color_from_int_passes_for_valid_values(self, i):
        assert colors.Color.from_int(i) is not None

    @pytest.mark.parametrize("i", [-1, 0x1000000])
    def test_Color_from_int_fails_for_out_of_range_values(self, i):
        with pytest.raises(ValueError, match=r"raw_rgb must be in the exclusive range of 0 and 16777215"):
            colors.Color.from_int(i)

    def test_equality_with_int(self):
        assert colors.Color(0xFA) == 0xFA

    def test_cast_to_int(self):
        assert int(colors.Color(0xFA)) == 0xFA

    @pytest.mark.parametrize(
        ("i", "string"), [(0x1A2B3C, "Color(r=0x1a, g=0x2b, b=0x3c)"), (0x1A2, "Color(r=0x0, g=0x1, b=0xa2)")]
    )
    def test_Color_repr_operator(self, i, string):
        assert repr(colors.Color(i)) == string

    @pytest.mark.parametrize(("i", "string"), [(0x1A2B3C, "#1A2B3C"), (0x1A2, "#0001A2")])
    def test_Color_str_operator(self, i, string):
        assert str(colors.Color(i)) == string

    @pytest.mark.parametrize(("i", "string"), [(0x1A2B3C, "#1A2B3C"), (0x1A2, "#0001A2")])
    def test_Color_hex_code(self, i, string):
        assert colors.Color(i).hex_code == string

    @pytest.mark.parametrize(("i", "string"), [(0x1A2B3C, "1A2B3C"), (0x1A2, "0001A2")])
    def test_Color_raw_hex_code(self, i, string):
        assert colors.Color(i).raw_hex_code == string

    @pytest.mark.parametrize(
        ("i", "expected_outcome"), [(0x1A2B3C, False), (0x1AAA2B, False), (0x0, True), (0x11AA33, True)]
    )
    def test_Color_is_web_safe(self, i, expected_outcome):
        assert colors.Color(i).is_web_safe is expected_outcome

    @pytest.mark.parametrize(("r", "g", "b", "expected"), [(0x9, 0x18, 0x27, 0x91827), (0x55, 0x1A, 0xFF, 0x551AFF)])
    def test_Color_from_rgb(self, r, g, b, expected):
        assert colors.Color.from_rgb(r, g, b) == expected

    def test_color_from_rgb_raises_value_error_on_invalid_red(self):
        with pytest.raises(ValueError, match=r"Expected red channel to be in the inclusive range of 0 and 255"):
            colors.Color.from_rgb(0x999, 32, 32)

    def test_color_from_rgb_raises_value_error_on_invalid_green(self):
        with pytest.raises(ValueError, match=r"Expected green channel to be in the inclusive range of 0 and 255"):
            colors.Color.from_rgb(32, 0x999, 32)

    def test_color_from_rgb_raises_value_error_on_invalid_blue(self):
        with pytest.raises(ValueError, match=r"Expected blue channel to be in the inclusive range of 0 and 255"):
            colors.Color.from_rgb(32, 32, 0x999)

    @pytest.mark.parametrize(
        ("r", "g", "b", "expected"),
        [(0x09 / 0xFF, 0x18 / 0xFF, 0x27 / 0xFF, 0x91827), (0x55 / 0xFF, 0x1A / 0xFF, 0xFF / 0xFF, 0x551AFF)],
    )
    def test_Color_from_rgb_float(self, r, g, b, expected):
        assert math.isclose(colors.Color.from_rgb_float(r, g, b), expected, abs_tol=1)

    def test_color_from_rgb_float_raises_value_error_on_invalid_red(self):
        with pytest.raises(ValueError, match=r"Expected red channel to be in the inclusive range of 0.0 and 1.0"):
            colors.Color.from_rgb_float(1.5, 0.5, 0.5)

    def test_color_from_rgb_float_raises_value_error_on_invalid_green(self):
        with pytest.raises(ValueError, match=r"Expected green channel to be in the inclusive range of 0.0 and 1.0"):
            colors.Color.from_rgb_float(0.5, 1.5, 0.5)

    def test_color_from_rgb_float_raises_value_error_on_invalid_blue(self):
        with pytest.raises(ValueError, match=r"Expected blue channel to be in the inclusive range of 0.0 and 1.0"):
            colors.Color.from_rgb_float(0.5, 0.5, 1.5)

    @pytest.mark.parametrize(("input", "r", "g", "b"), [(0x91827, 0x9, 0x18, 0x27), (0x551AFF, 0x55, 0x1A, 0xFF)])
    def test_Color_rgb(self, input, r, g, b):
        assert colors.Color(input).rgb == (r, g, b)

    @pytest.mark.parametrize(
        ("input", "r", "g", "b"),
        [(0x91827, 0x09 / 0xFF, 0x18 / 0xFF, 0x27 / 0xFF), (0x551AFF, 0x55 / 0xFF, 0x1A / 0xFF, 0xFF / 0xFF)],
    )
    def test_Color_rgb_float(self, input, r, g, b):
        assert colors.Color(input).rgb_float == (r, g, b)

    @pytest.mark.parametrize("prefix", ["0x", "0X", "#", ""])
    @pytest.mark.parametrize(
        ("expected", "string"), [(0x1A2B3C, "1A2B3C"), (0x1A2, "0001A2"), (0xAABBCC, "ABC"), (0x00AA00, "0A0")]
    )
    def test_Color_from_hex_code(self, prefix, string, expected):
        actual_string = prefix + string
        assert colors.Color.from_hex_code(actual_string) == expected

    def test_Color_from_hex_code_ValueError_when_not_hex(self):
        with pytest.raises(ValueError, match=r"Color code must be hexadecimal"):
            colors.Color.from_hex_code("0xlmfao")

    def test_Color_from_hex_code_ValueError_when_not_6_or_3_in_size(self):
        with pytest.raises(ValueError, match=r"Color code is invalid length. Must be 3 or 6 digits"):
            colors.Color.from_hex_code("0x1111")

    def test_Color_from_bytes(self):
        assert colors.Color(0xFFAAFF) == colors.Color.from_bytes(b"\xff\xaa\xff\x00\x00\x00\x00\x00\x00\x00", "little")

    def test_Color_to_bytes(self):
        c = colors.Color(0xFFAAFF)
        b = c.to_bytes(10, "little")
        assert b == b"\xff\xaa\xff\x00\x00\x00\x00\x00\x00\x00"

    @pytest.mark.parametrize(
        ("input", "expected_result"),
        [
            (colors.Color(0xFF051A), colors.Color(0xFF051A)),
            (0xFF051A, colors.Color(0xFF051A)),
            ((1.0, 0.5, 0.0), colors.Color(0xFF7F00)),
            ([0xFF, 0x5, 0x1A], colors.Color(0xFF051A)),
            ("#1a2b3c", colors.Color(0x1A2B3C)),
            ("#123", colors.Color(0x112233)),
            ("0x1a2b3c", colors.Color(0x1A2B3C)),
            ("0x123", colors.Color(0x112233)),
            ("0X1a2b3c", colors.Color(0x1A2B3C)),
            ("0X123", colors.Color(0x112233)),
            ("1a2b3c", colors.Color(0x1A2B3C)),
            ("123", colors.Color(0x112233)),
            ((1.0, 0.0196078431372549, 0.10196078431372549), colors.Color(0xFF051A)),
            ([1.0, 0.0196078431372549, 0.10196078431372549], colors.Color(0xFF051A)),
            *tuple_str_happy_test_data,
        ],
    )
    def test_Color_of_happy_path(self, input, expected_result):
        result = colors.Color.of(input)
        assert result == expected_result, f"{input}"

    @pytest.mark.parametrize(
        ("input_string", "value_error_match"),
        [
            ("blah", r"Could not transform 'blah' into a Color object"),
            ("0xfff1", r"Color code is invalid length\. Must be 3 or 6 digits"),
            (lambda: 22, r"Could not transform <function TestColor\.<lambda> at 0x[a-zA-Z0-9]+> into a Color object"),
            (NotImplementedError, r"Could not transform <class 'NotImplementedError'> into a Color object"),
            (NotImplemented, r"Could not transform NotImplemented into a Color object"),
            ((1, 1, 1, 1), r"Color must be an RGB triplet if set to a tuple type"),
            ((1, "a", 1), r"Could not transform \(1, 'a', 1\) into a Color object"),
            ((1.1, 1.0, 1.0), r"Expected red channel to be in the inclusive range of 0.0 and 1.0"),
            ((1.0, 1.1, 1.0), r"Expected green channel to be in the inclusive range of 0.0 and 1.0"),
            ((1.0, 1.0, 1.1), r"Expected blue channel to be in the inclusive range of 0.0 and 1.0"),
            ((), r"Color must be an RGB triplet if set to a tuple type"),
            ({}, r"Could not transform \{\} into a Color object"),
            ([], r"Color must be an RGB triplet if set to a list type"),
            ({1, 1, 1}, r"Could not transform \{1\} into a Color object"),
            (set(), r"Could not transform set\(\) into a Color object"),
            (b"1ff1ff", r"Could not transform b'1ff1ff' into a Color object"),
            *tuple_str_sad_test_data,
        ],
    )
    def test_Color_of_sad_path(self, input_string, value_error_match):
        with pytest.raises(ValueError, match=value_error_match):
            colors.Color.of(input_string)

    def test_Color_of_with_multiple_args(self):
        result = colors.Color.of((0xFF, 0x5, 0x1A))
        assert result == colors.Color(0xFF051A)

    @pytest.mark.parametrize(("input_string", "expected_color"), tuple_str_happy_test_data)
    def test_from_tuple_string_happy_path(self, input_string, expected_color):
        assert colors.Color.from_tuple_string(input_string) == expected_color

    @pytest.mark.parametrize(("input_string", "value_error_match"), tuple_str_sad_test_data)
    def test_from_tuple_string_sad_path(self, input_string, value_error_match):
        with pytest.raises(ValueError, match=value_error_match):
            colors.Color.from_tuple_string(input_string)



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Pyright integrations."""

from __future__ import annotations

from pipelines import config
from pipelines import nox


@nox.session()
def pyright(session: nox.Session) -> None:
    """Perform static type analysis on Python source code using pyright.

    At the time of writing this, this pipeline will not run successfully,
    as hikari does not have 100% compatibility with pyright just yet. This
    exists to make it easier to test and eventually reach that 100% compatibility.
    """
    nox.sync(session, self=True, extras=["speedups", "server"], groups=["pyright"])
    session.run("pyright")


@nox.session()
def verify_types(session: nox.Session) -> None:
    """Verify the "type completeness" of types exported by the library using Pyright."""
    nox.sync(session, self=True, groups=["pyright"])
    session.run("pyright", "--verifytypes", config.MAIN_PACKAGE, "--ignoreexternal")



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import mock
import pytest

from hikari import scheduled_events
from hikari import snowflakes
from hikari import urls
from hikari.internal import routes
from tests.hikari import hikari_test_helpers


class TestScheduledEvent:
    @pytest.fixture
    def model(self) -> scheduled_events.ScheduledEvent:
        return hikari_test_helpers.mock_class_namespace(scheduled_events.ScheduledEvent, init_=False, slots_=False)()

    def test_make_image_url_format_set_to_deprecated_ext_argument_if_provided(self, model):
        model.id = snowflakes.Snowflake(543123)
        model.image_hash = "ododododo"

        with mock.patch.object(
            routes, "SCHEDULED_EVENT_COVER", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_image_url(ext="JPEG") == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, scheduled_event_id=543123, hash="ododododo", size=4096, file_format="JPEG", lossless=True
        )

    def test_image_url_property(self, model: scheduled_events.ScheduledEvent):
        model.make_image_url = mock.Mock()

        assert model.image_url == model.make_image_url.return_value

        model.make_image_url.assert_called_once_with()

    def test_image_url(self, model: scheduled_events.ScheduledEvent):
        model.id = snowflakes.Snowflake(543123)
        model.image_hash = "ododododo"
        with mock.patch.object(routes, "SCHEDULED_EVENT_COVER") as route:
            assert model.make_image_url(file_format="JPEG", size=1) is route.compile_to_file.return_value

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, scheduled_event_id=543123, hash="ododododo", size=1, file_format="JPEG", lossless=True
        )

    def test_make_image_url_when_image_hash_is_none(self, model: scheduled_events.ScheduledEvent):
        model.image_hash = None

        with mock.patch.object(routes, "SCHEDULED_EVENT_COVER") as route:
            assert model.make_image_url(ext="JPEG", size=1) is None

        route.compile_to_file.assert_not_called()



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Enums use a lot of internal voodoo that will not type check nicely, so we
# skip that module with MyPy and just accept that "here be dragons".
#
# The caveat to implementing this is that MyPy has to have a special module to
# understand how to use Python's enum types. I really don't want to have to
# ship my own MyPy plugin for this, so just make MyPy think that the types
# we are using are just aliases from the enum types in the standard library.

__all__ = ["Enum", "Flag"]

import enum as __enum
from collections.abc import Iterator as __Iterator
from collections.abc import Sequence as __Sequence
from typing import TypeVar as __TypeVar

from typing_extensions import Self as __Self

from hikari.internal import typing_extensions as __typing_backport

Enum = __enum.Enum

# MyPy started complaining of Flags with no additional flags, so just ignore it here
class Flag(__enum.IntFlag):  # type: ignore[misc]
    def all(self, *flags: __Self) -> bool: ...
    def any(self, *flags: __Self) -> bool: ...
    def difference(self, other: int | __Self) -> __Self: ...
    def intersection(self, other: int | __Self) -> __Self: ...
    def invert(self) -> __Self: ...
    def is_disjoint(self, other: int | __Self) -> bool: ...
    def is_subset(self, other: int | __Self) -> bool: ...
    def is_superset(self, other: int | __Self) -> bool: ...
    def none(self, *flags: __Self) -> bool: ...
    def split(self) -> __Sequence[__Self]: ...
    def symmetric_difference(self, other: int | __Self) -> __Self: ...
    def union(self, other: int | __Self) -> __Self: ...
    def __iter__(self) -> __Iterator[__Self]: ...
    def __len__(self) -> int: ...
    # Aliases
    def isdisjoint(self, other: int | __Self) -> bool: ...  # is_disjoint
    def issuperset(self, other: int | __Self) -> bool: ...  # is_superset
    def symmetricdifference(self, other: int | __Self) -> __Self: ...  # symmetric_difference
    def issubset(self, other: int | __Self) -> bool: ...  # is_subset
    # '__invert__' is explicitly defined as a special case because it is typed as returning 'int' in typeshed
    @__typing_backport.override
    def __invert__(self) -> __Self: ...  # invert

_V = __TypeVar("_V")

def deprecated(value: _V, /, *, removal_version: str) -> _V: ...
